#       Copyright (C) 2005-2007 Carabos Coop. V. All rights reserved
#       Copyright (C) 2008-2019 Vicent Mas. All rights reserved
#
#       This program is free software: you can redistribute it and/or modify
#       it under the terms of the GNU General Public License as published by
#       the Free Software Foundation, either version 3 of the License, or
#       (at your option) any later version.
#
#       This program is distributed in the hope that it will be useful,
#       but WITHOUT ANY WARRANTY; without even the implied warranty of
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#       GNU General Public License for more details.
#
#       You should have received a copy of the GNU General Public License
#       along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#       Author:  Vicent Mas - vmas@vitables.org

"""
This module defines a model (in the `MVC` sense) representing the tree of
databases.

The model is populated using data structures defined in the
:mod:`vitables.h5db.rootgroupnode`, :mod:`vitables.h5db.groupnode` and
:mod:`vitables.h5db.leafnode` modules.
"""

__docformat__ = 'restructuredtext'

import tempfile
import os
import sys
import re
import logging

import tables

from qtpy import QtCore
from qtpy import QtGui
from qtpy import QtWidgets

import vitables.utils
from vitables.h5db import dbdoc
from vitables.h5db import rootgroupnode
from vitables.h5db import groupnode
from vitables.h5db import leafnode
from vitables.h5db import linknode
from vitables.h5db import tnode_editor
from vitables.h5db import tlink_editor

translate = QtWidgets.QApplication.translate

log = logging.getLogger(__name__)


def _get_node_tooltip(node):
    """Takes one of vitables nodes and return tooltip text.

    :Parameters node: a node from the tree model
    """

    tooltip_lines = []
    attrset = node.node._v_attrs
    if not hasattr(node.node, 'target'):
        if hasattr(attrset, 'TITLE') and bool(str(attrset.TITLE)):
            tooltip_lines.append(attrset.TITLE)
    tooltip_lines.append('{0}: {1}'.format(node.node_kind, node.name))
    tooltip_lines.extend(
        ['{0}: {1}'.format(name, str(getattr(attrset, name)))
         for name in attrset._v_attrnamesuser])
    return '\n'.join(tooltip_lines)


# Map qt display role onto a function that gets data from a node.
_ROLE_NODE_GET_FUNCTION_DICT = {
    QtCore.Qt.DisplayRole: lambda n: getattr(n, 'name'),
    QtCore.Qt.ToolTipRole: _get_node_tooltip,
    QtCore.Qt.StatusTipRole: lambda n: getattr(n, 'as_record'),
    QtCore.Qt.DecorationRole: lambda n: getattr(n, 'icon'),
    QtCore.Qt.UserRole: lambda n: getattr(n, 'filepath'),
    QtCore.Qt.UserRole + 1: lambda n: getattr(n, 'nodepath'),
    QtCore.Qt.UserRole + 2: lambda n: getattr(n, 'node_kind')
}


class DBsTreeModel(QtCore.QAbstractItemModel):
    """
    The tree of databases model.

    The data is read and written from and to data sources (i.e., HDF5/PyTables
    files) by the model.

    :Parameters vtapp: the VTAPP instance
    """

    def __init__(self, vtgui, vtapp):
        """Create the model.
        """

        # The underlying data structure used to populate the model
        self.root = rootgroupnode.RootGroupNode(self)

        super(DBsTreeModel, self).__init__(parent=None)
        self.setObjectName("dbs_tree_model")

        # The dictionary of open databases
        self.__openDBs = {}

        # Create the temporary database that will contain filtered tables
        self.tmp_filepath = ''
        self.tmp_dbdoc = self.__createTempDB()

        # The Cut/CopiedNodeInfo dictionary
        self.ccni = {}
        self.vtapp = vtapp
        self.vtgui = vtgui

        # Sets used to populate the model
        self.fdelta = frozenset([])
        self.gdelta = frozenset([])
        self.ldelta = frozenset([])
        self.links_delta = frozenset([])

        self.rowsAboutToBeRemoved.connect(self.closeViews)

    def mapDB(self, filepath, db_doc):
        """Maps a file path with a :meth:`dbdoc.DBDoc` instance.

        :Parameters:

        - `filepath`: the full path of an open database.
        - `db_doc`: a :meth:`dbdoc.DBDoc` instance.
        """
        self.__openDBs[filepath] = db_doc

    def removeMappedDB(self, filepath):
        """Remove a :meth:`dbdoc.DBDoc` instance from the tracking dict.

        :Parameter filepath: the full path of the database being untracked
        """
        del self.__openDBs[filepath]

    def getDBDoc(self, filepath):
        """Returns the DBDoc instance tied to a given file path.

        :Parameter filepath: the full path of an open database.
        """
        if filepath in self.__openDBs:
            return self.__openDBs[filepath]
        else:
            return None

    def getDBList(self):
        """Returns the list of paths of files currently open.

        The temporary database is not included in the list.
        """

        filepaths = list(self.__openDBs.keys())
        try:
            filepaths.remove(self.tmp_filepath)
        except ValueError:
            pass
        return filepaths

    def checkOpening(self, filepath):
        """
        Check if a database can be open.

        :Parameter filepath: the full path of the file
        """
        try:
            # Check if file doesn't exist
            if os.path.isdir(filepath):
                error = translate('DBsTreeModel',
                                  'Openning cancelled: {0} is a folder.',
                                  'A logger error message').format(filepath)
                raise ValueError

            elif not os.path.isfile(filepath):
                error = translate('DBsTreeModel',
                                  'Opening failed: file {0} cannot be found.',
                                  'A logger error message').format(filepath)
                raise ValueError

            # Check if file is already open.
            elif self.getDBDoc(filepath) is not None:
                error = translate('DBsTreeModel',
                                  'Opening cancelled: file {0} already open.',
                                  'A logger error message').format(filepath)

                raise ValueError

        except ValueError:
            log.error(error)
            return False

        # Check the file format
        try:
            if not tables.is_hdf5_file(filepath):
                error = translate('DBsTreeModel',
                                  'Opening cancelled: file {0} has not HDF5 format.',
                                  'A logger error message').format(filepath)
                log.error(error)
                return False
        except (tables.NodeError, OSError):
            error = translate('DBsTreeModel',
                              """Opening failed: I cannot find out if file {0} has HDF5 """
                              """format.""",
                              'A logger error message').format(filepath)
            log.error(error)
            return False
        else:
            return True

    def openDBDoc(self, filepath, mode='a', position=0):
        """
        Open an existing hdf5 file and load it into the tree model.

        :Parameters:

        - `filepath`: full path of the database file we wish to open.
        - `mode`: the opening mode of the database file. It can be 'r'ead-only
          'w'rite or 'a'ppend
        """
        is_open = False
        if self.checkOpening(filepath):
            # Open the database and add it to the tracking system
            db_doc = dbdoc.DBDoc(filepath, mode)
            self.mapDB(filepath, db_doc)

            # Populate the model with the dbdoc
            root_node = rootgroupnode.RootGroupNode(self, db_doc, self.root)
            self.fdelta = frozenset([root_node])
            self.gdelta = frozenset([])
            self.ldelta = frozenset([])
            self.links_delta = frozenset([])
            self.insertRows(position, 1)
            is_open = True

        return is_open

    def closeDBDoc(self, filepath):
        """
        Close the hdf5 file with the given file path.

        The temporary database can't be closed by the user. It is
        automatically closed when the application exits.

        :Parameter filepath: the full path of the file being closed
        """

        for row, child in enumerate(self.root.children):
            if child.filepath == filepath:
                # Deletes the node from the tree of databases model/view
                self.remove_rows(row)

                # If needed, refresh the copied node info
                try:
                    if self.ccni['filepath'] == filepath:
                        self.ccni = {}
                except KeyError:
                    pass

                # Close the hdf5 file
                db_doc = self.getDBDoc(filepath)
                if db_doc.hidden_group is not None:
                    db_doc.h5file.remove_node(db_doc.hidden_group,
                                              recursive=True)
                db_doc.closeH5File()
                # Update the dictionary of open files
                self.removeMappedDB(filepath)
                break

    def createDBDoc(self, filepath, is_tmp_db=False):
        """
        Create a new, empty database (:meth:`vitables.h5db.dbdoc.DBDoc`
        instance).

        :Parameters:

        - `filepath`: the full path of the file being created.
        - `is_tmp_db`: True if the `DBDoc` is tied to the temporary database
        """

        try:
            QtWidgets.qApp.setOverrideCursor(
                QtGui.QCursor(QtCore.Qt.WaitCursor))
            # Create the dbdoc
            try:
                db_doc = dbdoc.DBDoc(filepath, 'w', is_tmp_db)
            except (tables.NodeError, OSError):
                log.error(
                    translate('DBsTreeModel',
                              """File creation failed due to unknown"""
                              """reasons! Please, have a look to the """
                              """last error displayed in the logger. If you """
                              """think it's a bug, please report it to """
                              """developers.""", 'A file creation error'))
                return None

            # Track the just created dbdoc
            self.mapDB(filepath, db_doc)

            # Populate the model with the dbdoc
            root = rootgroupnode.RootGroupNode(self, db_doc, self.root,
                                               is_tmp_db)
            self.fdelta = frozenset([root])
            self.gdelta = frozenset([])
            self.ldelta = frozenset([])
            self.links_delta = frozenset([])
            self.insertRows(0, 1)
        finally:
            QtWidgets.qApp.restoreOverrideCursor()
        return db_doc

    def __createTempDB(self):
        """
        Create a temporary database where filtered tables will be stored.

        The database will have a hidden group where cut nodes can be
        stored until they are pasted somewhere else.
        """

        # Create the database
        log.info(
            translate('DBsTreeModel', 'Creating the Query results file...',
                      'A logger info message'))
        (f_handler, filepath) = tempfile.mkstemp('.h5', 'FT_')
        os.close(f_handler)
        self.tmp_filepath = vitables.utils.forwardPath(filepath)
        db_doc = self.createDBDoc(self.tmp_filepath, True)
        return db_doc

    def deleteNode(self, index):
        """Delete a node.

        Delete the selected node from the model and from the database
        where it lives.

        :Parameter index: the index of the selected node
        """

        try:
            QtWidgets.qApp.setOverrideCursor(QtCore.Qt.WaitCursor)
            node = self.nodeFromIndex(index)
            # Deletes the node from the database
            node.editor().delete(node.nodepath)

            # Deletes the node from the tree of databases model/view
            parent = self.parent(index)
            position = node.row()
            self.remove_rows(position, parent=parent)

            # If needed, refresh the copied node info. This info must
            # not be cleared when pasting a CUT node overwrites other
            # node
            try:
                if self.ccni['is_copied'] and \
                        (self.ccni['filepath'] == node.filepath) and \
                        self.ccni['nodepath'].startswith(node.nodepath):
                    self.ccni = {}
            except KeyError:
                pass
        finally:
            QtWidgets.qApp.restoreOverrideCursor()

    def copy_node(self, index):
        """Mark a node from the tree of databases view as copied.

        :Parameter index: the index of the selected node
        """

        try:
            QtWidgets.qApp.setOverrideCursor(QtCore.Qt.WaitCursor)
            node = self.nodeFromIndex(index)
            self.ccni = {'is_copied': True,
                         'nodename': node.name,
                         'filepath': node.filepath,
                         'nodepath': node.nodepath,
                         'target': getattr(node, 'target', None)}
        finally:
            QtWidgets.qApp.restoreOverrideCursor()

    def cutNode(self, index):
        """Cut a `tables.Node`.

        The cut node is stored in a hidden group of its database.

        :Parameter index: the index of the selected node
        """

        try:
            QtWidgets.qApp.setOverrideCursor(QtCore.Qt.WaitCursor)
            node = self.nodeFromIndex(index)
            self.ccni = {'is_copied': False,
                         'nodename': node.name,
                         'filepath': node.filepath,
                         'nodepath': node.nodepath,
                         'target': getattr(node, 'target', None)}

            # Moves the node to a hidden group in its database
            node.editor().cut(node.nodepath)

            # Deletes the node from the tree of databases model/view
            parent = self.parent(index)
            position = index.row()
            self.remove_rows(position, parent=parent)
            # If position > 0 then the above sibling will be selected
            # else the parent group will be selected
            if position == 0:
                current = parent
            else:
                current = self.index(position - 1, 0, parent)
            self.vtgui.dbs_tree_view.selectNode(current)
        finally:
            QtWidgets.qApp.restoreOverrideCursor()

    def pasteNode(self, index, childname, overwrite=False):
        """Paste a tables.Node.

        Paste the last copied/cut node under the currently selected group.

        :Parameters:

        - `index`: the index of the selected node (the parent group)
        - `childname`: the name of the node being pasted
        - `overwrite`: True if a node is being overwritten
        """

        try:
            QtWidgets.qApp.setOverrideCursor(QtCore.Qt.WaitCursor)
            parent = self.nodeFromIndex(index)

            # If the overwritten node (if any) exists in the tree of
            # databases view then delete it
            if overwrite:
                self.overwriteNode(parent, index, childname)

            # Paste the copied/cut node in the destination database
            db_doc = self.getDBDoc(parent.filepath)
            if self.ccni['target']:
                editor = tlink_editor.TLinkEditor(db_doc)
            else:
                editor = tnode_editor.TNodeEditor(db_doc)

            src_node = self.copiedNode()
            editor.paste(src_node, parent.node, childname)

            # Paste the node in the tree of databases model/view
            self.lazyAddChildren(index)
            parent.updated = True

            # Select the pasted node
            self.selectIndex(index, childname)
        finally:
            QtWidgets.qApp.restoreOverrideCursor()

    def copiedNode(self):
        """The tables.Node currently copied/cut.
        """

        src_filepath = self.ccni['filepath']
        if self.ccni['is_copied']:
            src_nodepath = self.ccni['nodepath']
        else:
            dirname = self.getDBDoc(src_filepath).hidden_group
            basename = self.ccni['nodename']
            src_nodepath = '{0}/{1}'.format(dirname, basename)

        return self.getDBDoc(src_filepath).get_node(src_nodepath)

    def create_group(self, index, childname, overwrite=False):
        """Create a `tables.Group` under the given parent.

        :Parameters:

        - `index`: the index of the parent node
        - `childname`: the name of the group being created
        - `overwrite`: True if a node is being overwritten
        """

        try:
            QtWidgets.qApp.setOverrideCursor(QtCore.Qt.WaitCursor)
            parent = self.nodeFromIndex(index)
            # If the overwritten node (if any) exists in the tree of
            # databases view then delete it
            if overwrite:
                self.overwriteNode(parent, index, childname)

            # Create the group in the PyTables database
            parent.editor().create_group(parent.nodepath, childname)

            # Paste the node in the tree of databases model/view
            self.lazyAddChildren(index)
            parent.updated = True

            # Select the pasted node
            self.selectIndex(index, childname)
        finally:
            QtWidgets.qApp.restoreOverrideCursor()

    def rename_node(self, index, new_name, overwrite=False):
        """Rename a node.

        :Parameters:

        - `index`: the index of the node being renamed
        - `new_name`: the new name of the node
        - `overwrite`: True if a node is being overwritten
        """

        try:
            QtWidgets.qApp.setOverrideCursor(QtCore.Qt.WaitCursor)
            node = self.nodeFromIndex(index)
            initial_nodepath = node.nodepath
            parent_index = self.parent(index)
            parent = self.nodeFromIndex(parent_index)
            if overwrite:
                self.overwriteNode(parent, parent_index, new_name)

            # Rename the node in the PyTables database
            node.editor().rename(node.nodepath, new_name)

            # Rename the node in the databases tree view
            # The renamed node's children must be updated too
            self.updateDBTree(index, new_name, node)

            # If needed, refresh the copied node info
            try:
                if self.ccni['filepath'] == node.filepath:
                    if self.ccni['nodepath'].startswith(initial_nodepath):
                        self.ccni = {}
            except KeyError:
                pass
        finally:
            QtWidgets.qApp.restoreOverrideCursor()

    def updateDBTree(self, index, new_name, node):
        """
        After renaming a PyTables node update the tree of databases model/view.

        The following data must be updated:

            - name (DisplayRole)
            - nodepath (UserRole+1)
            - status tip (StatusTipRole)

        :Parameters:

        - `index`: the index of the node being renamed
        - `new_name`: the new name of the node
        - `node`: the node being renamed
        """

        # Update the renamed tree item
        self.setData(index, new_name, QtCore.Qt.DisplayRole)
        old_nodepath = node.nodepath
        dirname = os.path.split(old_nodepath)[0]
        new_nodepath = \
            ('{0}/{1}'.format(dirname, new_name)).replace('//', '/')
        self.setData(index, new_nodepath, QtCore.Qt.UserRole + 1)
        if hasattr(node, 'target'):
            self.setData(index, '{0}'.format(node.node),
                         QtCore.Qt.StatusTipRole)
        else:
            self.setData(index,
                         '{0}->{1}'.format(node.filepath, new_nodepath),
                         QtCore.Qt.StatusTipRole)

        # Update the item children, if any
        for child_index in self.walkTreeView(index):
            child_node = self.nodeFromIndex(child_index)
            child_nodepath = child_node.nodepath.replace(old_nodepath,
                                                         new_nodepath, 1)
            self.setData(child_index, child_nodepath, QtCore.Qt.UserRole + 1)
            if hasattr(child_node, 'target'):
                self.setData(child_index, '{0}'.format(child_node.node),
                             QtCore.Qt.StatusTipRole)
            else:
                self.setData(child_index,
                             '{0}->{1}'.format(child_node.filepath,
                                               child_node.nodepath),
                             QtCore.Qt.StatusTipRole)

    def move_node(self, src_filepath, childpath, parent_index, overwrite=False):
        """Move a `tables.Node` to a different location.

        :Parameters:

          - `src_filepath`: the full path of the source database
          - `childpath`: the full path of the node being moved
          - `parent_index`: the model index of the new parent group
          - `overwrite`: True if a node is being overwritten
        """

        try:
            QtWidgets.qApp.setOverrideCursor(QtCore.Qt.WaitCursor)
            parent_node = self.nodeFromIndex(parent_index)
            # full path of the destination database and new parent
            dst_filepath = parent_node.filepath
            parentpath = parent_node.nodepath

            #
            # Check if the nodename is already in use
            #
            (nodename, overwrite) = self.validateNodename(
                src_filepath, childpath, dst_filepath, parentpath)
            if nodename is None:
                return nodename

            # If the overwritten node (if any) exists in the tree of
            # databases view then delete it
            if overwrite:
                self.overwriteNode(parent_node, parent_index, nodename)

            # Move the node to the PyTables database
            pt_node = self.getDBDoc(src_filepath).get_node(childpath)
            db_doc = self.getDBDoc(src_filepath)
            if hasattr(pt_node, 'target'):
                editor = tlink_editor.TLinkEditor(db_doc)
            else:
                editor = tnode_editor.TNodeEditor(db_doc)
            movedname = editor.move(childpath, self.getDBDoc(dst_filepath),
                                    parentpath, nodename)
        finally:
            QtWidgets.qApp.restoreOverrideCursor()
            return movedname

    def validateNodename(self, src_filepath, childpath, dst_filepath,
                         parentpath):
        """

        :Parameters:

          - `childpath`: the full path of the node being moved
          - `dst_filepath`: the full path of the destination database
          - `parentpath`: the full path of the new parent group
        """

        nodename = os.path.basename(childpath)
        dst_dbdoc = self.getDBDoc(dst_filepath)
        parent = dst_dbdoc.get_node(parentpath)
        sibling = getattr(parent, '_v_children').keys()

        # Nodename pattern
        pattern = "[a-zA-Z_]+[0-9a-zA-Z_ ]*"
        info = [translate('DBsTreeModel',
                          'Node move: nodename already exists',
                          'A dialog caption'),
                translate('DBsTreeModel',
                          """Source file: {0}\nMoved node: {1}\n"""
                          """Destination file: {2}\nParent group: {3}\n\n"""
                          """Node name '{4}' already in use in that group.\n""",
                          'A dialog label').format
                (src_filepath, childpath, dst_filepath,
                    parentpath, nodename),
                translate('DBsTreeModel', 'Rename', 'A button label')]

        # Validate the nodename
        nodename, overwrite = vitables.utils.getFinalName(
            nodename, sibling, pattern, info)
        self.vtgui.editing_dlg = True
        return nodename, overwrite

    def overwriteNode(self, parent_node, parent_index, nodename):
        """Delete from the tree of databases a node being overwritten.

        :Parameters:

          - `parent_node`: the parent of the overwritten node
          - `parent_index`: the model index of the new parent group
          - `nodename`: the name of the node being deleted
        """

        child = parent_node.findChild(nodename)
        child_index = self.index(child.row(), 0, parent_index)
        self.deleteNode(child_index)

    def walkTreeView(self, index):
        """Iterates over a subtree of the tree of databases view.

        :Parameter index: the model index of the root node of the iterated
            subtree
        """

        for prow in range(0, self.rowCount(index)):
            child = self.index(prow, 0, index)
            yield child
            for crow in range(0, self.rowCount(child)):
                yield self.index(crow, 0, child)

    def indexChildren(self, index):
        """Iterate over the children of a given index.

        :Parameter index: the model index whose children are being retrieved
        """

        for row in range(0, self.rowCount(index)):
            yield self.index(row, 0, index)

    def selectIndex(self, parent, name):
        """Select in the tree view the index with the given parent and name.

        :Parameters:

            - `parent`: the parent of the model index being selected
            - `name`: the name tied to the model index being selected
        """

        for child in self.indexChildren(parent):
            node = self.nodeFromIndex(child)
            if node.name == name:
                self.vtgui.dbs_tree_view.selectNode(child)

    def updateTreeFromData(self, parent_index=None):
        """Update tree of expanded nodes from hdf5 data.

        This function should be used if the structure of a file was
        updated through pytables functions. It is better to use model
        functions but they need to be better documented and expanded
        to be usefull. This function is a temporary fix.

        """
        parent_index = parent_index if parent_index else QtCore.QModelIndex()
        self.lazyAddChildren(parent_index)
        for index in self.indexChildren(parent_index):
            self.updateTreeFromData(index)

    def flags(self, index):
        """Returns the item flags for the given index.

        This is a reimplemented function used by the model to indicate
        to views which items can be dragged, and which will accept drops.

        :Parameter index: the index of a given item.
        """

        # Every item in the model is enabled and selectable
        default_flags = QtCore.Qt.ItemIsEnabled | QtCore.Qt.ItemIsSelectable

        if not index.isValid():
            return default_flags

        node = self.nodeFromIndex(index)
        node_kind = node.node_kind
        # If a database is in read-only mode items cannot be edited
        db_doc = self.getDBDoc(node.filepath)
        if db_doc.mode == 'r':
            flags = default_flags
        # The root group of a database cannot be moved or renamed
        elif node_kind == 'root group':
            flags = QtCore.Qt.ItemIsDropEnabled | default_flags
        # A regular group of a database can be moved, renamed and dropped
        elif node_kind == 'group':
            flags = QtCore.Qt.ItemIsDragEnabled | QtCore.Qt.ItemIsDropEnabled \
                | QtCore.Qt.ItemIsEditable | default_flags
        # A leaf of a database can be moved and renamed but not dropped
        else:
            flags = QtCore.Qt.ItemIsDragEnabled | QtCore.Qt.ItemIsEditable \
                | default_flags

        return flags

    def data(self, index, role):
        """Returns the data stored under the given role for the item
        referred to by the index.

        :Parameters:

        - `index`: the index of a data item
        - `role`: the role being returned
        """
        if not index.isValid():
            data = None
            return data
        node = self.nodeFromIndex(index)
        if role in _ROLE_NODE_GET_FUNCTION_DICT:
            return _ROLE_NODE_GET_FUNCTION_DICT[role](node)
        else:
            return None

    def setData(self, index, value, role=QtCore.Qt.EditRole):
        """Sets the role data for the item at index to value.

        :Parameters:

        - `index`: the index of a data item
        - `value`: the value to be set
        - `role`: the role being set
        """

        if not index.isValid():
            result = False
        node = self.nodeFromIndex(index)
        if role == QtCore.Qt.DisplayRole:
            node.name = value
            self.dataChanged.emit(index, index)
            result = True
        elif role == QtCore.Qt.StatusTipRole:
            node.as_record = value
            self.dataChanged.emit(index, index)
            result = True
        elif role == QtCore.Qt.DecorationRole:
            node.icon = value
            self.dataChanged.emit(index, index)
            result = True
        elif role == QtCore.Qt.UserRole:
            node.filepath = value
            self.dataChanged.emit(index, index)
            result = True
        elif role == QtCore.Qt.UserRole + 1:
            node.nodepath = value
            self.dataChanged.emit(index, index)
            result = True
        elif role == QtCore.Qt.UserRole + 2:
            node.node_kind = value
            self.dataChanged.emit(index, index)
            result = True
        else:
            result = False
        return result

    def headerData(self, section, orientation, role):
        """Returns the data for the given role and section in the header
        with the specified orientation.

        :Parameters:

        - `section`: the header section being inspected
        - `orientation`: the header orientation (horizontal or vertical)
        - `role`: the role of the header section being inspected
        """

        if (orientation, role) == (QtCore.Qt.Horizontal,
                                   QtCore.Qt.DisplayRole):
            return translate('DBsTreeModel',
                             'Tree of databases',
                             'Header of the only column of the tree of databases view')

        return None

    def columnCount(self, index):
        """The number of columns for the children of the given index.

        :Parameter index: the model index being inspected.
        """
        return 1

    def rowCount(self, index):
        """The number of rows of the given model index.

        :Parameter index: the model index being inspected.
        """

        node = self.nodeFromIndex(index)
        if node == self.root:
            nrows = len(node)
        else:
            nrows = 0
            if node.node_kind in ('group', 'root group'):
                nrows = len(node)
        return nrows

    def hasChildren(self, index):
        """Finds out if a node has children.

        The goal of this reimplementation is to provide an inexpensive
        way for views to check for the presence of children and draw the
        appropriate decoration for their parent item.

        The decoration (if any) consists in the small +/- symbols used
        for expanding/collpasing the node. In principle, it is painted
        only if the node's children have been added.
        As we populate our model in a lazy way (see :meth:`lazyAddChildren`
        and :meth:`vitables.h5db.dbstreeview.DBsTreeView.updateExpandedGroup`
        methods) we want the decoration to be painted whenever the node has
        children, *even if the children have not been added to the model yet
        (so we can't use the underlying data store, we must use the data
        source). This way the user will know that the node has children*.

        :Parameter index: the index of the node being inspected.
        """

        if not index.isValid():  # self.root fulfils this condition
            return True
        parent = self.nodeFromIndex(index)
        if hasattr(parent.node, '_v_nchildren'):
            return getattr(parent.node, '_v_nchildren')
        else:
            return False

    def index(self, row, column, parent):
        """Creates an index in the model for a given node and returns it.

        This is a reimplementation of the index method. It creates an
        index for a tree node (i.e. :meth:`RootGroupNode`, :meth:`GroupNode`
        and :meth:`LeafNode` instances) specified by a row, a column and a
        parent index.

        Every node except the root one will be tied to an index by this
        method.

        :Parameters:

        - `row`: the row of the node relative to its parent
        - `column`: the column of the node relative to its parent
        - `parent`: the index of the node's parent
        """

        if self.hasIndex(row, column, parent):
            group = self.nodeFromIndex(parent)
            try:
                node = group.childAtRow(row)
                return self.createIndex(row, column, node)
            except IndexError:
                return QtCore.QModelIndex()
        return QtCore.QModelIndex()

    def nodeFromIndex(self, index):
        """Retrieves the tree node with a given index.

        :Parameter index: the index of a tree node
        """

        if index.isValid():
            return index.internalPointer()
        else:
            return self.root

    def parent(self, child):
        """The parent index of a given index.

        :Parameter child: the child index whose parent is being retrieved.
        """

        node = self.nodeFromIndex(child)
        if node is self.root:
            return QtCore.QModelIndex()
        parent = node.parent
        if parent is self.root:
            return QtCore.QModelIndex()
        grandparent = parent.parent
        row = grandparent.rowOfChild(parent)
        return self.createIndex(row, 0, parent)

    def lazyAddChildren(self, index):
        """Add children to a group node when it is expanded.

        Lazy population of the model is partially implemented in this
        method. Expanded items are updated so that children items are added if
        needed. This fact improves enormously the performance when files
        whit a large number of nodes are opened.

        :Parameter index: the index of the activated item
        """

        if not index.isValid():
            return
        node = self.nodeFromIndex(index)
        if node.node_kind not in ('group', 'root group'):
            return
        group = node.node

        # Find out if children have to be added by comparing the
        # names of children currently added to model with the
        # names of the whole list of children
        model_children = frozenset([node.childAtRow(row).name
                                    for row in range(0, len(node))])
        children_groups = frozenset(getattr(group, '_v_groups').keys())
        children_leaves = frozenset(getattr(group, '_v_leaves').keys())
        children_links = frozenset(getattr(group, '_v_links').keys())
        self.gdelta = children_groups.difference(model_children)
        self.ldelta = children_leaves.difference(model_children)
        self.links_delta = children_links.difference(model_children)
        self.fdelta = frozenset([])
        new_children = (len(self.gdelta) + len(self.ldelta)
                        + len(self.links_delta))
        if not new_children:
            return
        else:
            self.insertRows(0, new_children, index)

    def insertRows(self, position=0, count=1, parent=QtCore.QModelIndex()):
        """Insert `count` rows before the given `position`.

        Inserted rows will be children of the item represented by the `parent`
        model index. This method is called during nodes population and when
        files are opened/created.

        Warning! This method is MONKEY PATCHED if the DBs Sorting plugin is
        enabled. If it is the case the working code will be in module
        vitables.plugins.dbstreesort.dbs_tree_sort

        :Parameters:

        - `position`: the position of the first row being added.
        - `count`: the number of rows being added
        - `parent`: the index of the parent item.

        :Returns: True if the row is added. Otherwise it returns False.
        """

        # Add rows to the model and update its underlaying data store
        self.layoutAboutToBeChanged.emit()
        first = position
        last = position + count - 1
        self.beginInsertRows(parent, first, last)
        node = self.nodeFromIndex(parent)
        for file_node in self.fdelta:
            self.root.insertChild(file_node, position)
        for name in self.gdelta:
            group = groupnode.GroupNode(self, node, name)
            node.insertChild(group, position)
        for name in self.ldelta:
            leaf = leafnode.LeafNode(self, node, name)
            node.insertChild(leaf, position)
        for name in self.links_delta:
            link = linknode.LinkNode(self, node, name)
            node.insertChild(link, position)
        self.dataChanged.emit(parent, parent)
        self.endInsertRows()
        self.layoutChanged.emit()

        # Report views about changes in data
        top_left = self.index(first, 0, parent)
        bottom_right = self.index(last, 0, parent)
        self.dataChanged.emit(top_left, bottom_right)

        return True

    def remove_rows(self, position, count=1, parent=QtCore.QModelIndex()):
        """Removes `count` rows before the given row.

        This method is called when an item is cut/deleted/D&D.

        :Parameters:

        - `position`: the position of the first row being removed.
        - `count`: the number of rows being removed
        - `parent`: the index of the parent item.

        :Returns: True if the row is removed. Otherwise it returns False.
        """

        # Remove rows from the model and update its underlaying data store
        self.layoutAboutToBeChanged.emit()
        first = position
        last = position + count - 1
        self.beginRemoveRows(parent, first, last)
        group = self.nodeFromIndex(parent)
        del group.children[position]
        self.dataChanged.emit(parent, parent)
        self.endRemoveRows()
        self.layoutChanged.emit()

        return True

    def closeViews(self, parent, start, end):
        """When a leaf with a view is about to be removed then close the view.

        :Parameters:

            - `parent`: model index under which items are going to be removed
            - `start`: position of the first removed index
            - `stop`: position of the last removed index
        """

        nodepaths = []
        for position in range(start, end + 1):
            node = self.nodeFromIndex(self.index(position, 0, parent))
            nodepaths.append(node.nodepath)
        filepath = node.filepath
        for window in self.vtgui.workspace.subWindowList():
            if window.dbt_leaf.filepath == filepath:
                wpath = window.dbt_leaf.nodepath
                for path in nodepaths:
                    if re.match(path, wpath):
                        window.close()
                        break

    def supportedDropActions(self):
        """Setup drag and drop behavior of the model."""
        return QtCore.Qt.CopyAction | QtCore.Qt.MoveAction

    def mimeTypes(self):
        """Returns a list of MIME types that can be used to describe a
        list of model indexes or a list of filepaths.
        """
        return ["application/x-dbstreemodeldatalist", "text/uri-list"]

    def mimeData(self, indexes):
        """Returns a QMimeData object that contains serialized items of data
        corresponding to the list of indexes specified.

        When a node is dragged the information required to drop it later on is
        encoded by this method and returned as a QMimeData object.

        :Parameter indexes: a list of indexes
        """

        mime_data = QtCore.QMimeData()
        encoded_data = QtCore.QByteArray()

        stream = QtCore.QDataStream(encoded_data, QtCore.QIODevice.WriteOnly)

        # Only one item selections are allowed in the tree of databases
        # view so indexes contains always one element
        for index in indexes:
            if index.isValid():
                filepath = self.data(index, QtCore.Qt.UserRole)
                nodepath = self.data(index, QtCore.Qt.UserRole + 1)
                row = str(index.row())
                stream.writeQString(filepath)
                stream.writeQString(nodepath)
                stream.writeQString(row)

                self.initial_parent = self.parent(index)

        # Store the MIME type of the object and the encoded description of that
        # object in the QMimeData object
        mime_data.setData("application/x-dbstreemodeldatalist", encoded_data)
        return mime_data

    def dropMimeData(self, data, action, row, column, parent):
        """Handles the data supplied by a drag and drop operation that
        ended with the given action.

        :Parameters:

        - `data`: the data being dropped.
        - `action`: the action being performed (copy or move).
        - `row`: the row under the item where the operation ends.
        - `columns`: the column under the item where the operation ends.
        - `parent`: the index of the item where operation ends.

        :Returns: True if item is dropped. Otherwise it returns False.
        """

        # Examine the MIME type
        if not (data.hasFormat("application/x-dbstreemodeldatalist") or
                data.hasFormat("text/uri-list")):
            return False

        if action == QtCore.Qt.IgnoreAction:
            return True

        if column > 0:
            return False

        # If the dropped object is one or more files then open them
        if data.hasFormat("text/uri-list"):
            uris_list = data.urls()
            # Transform every element of the sequence into a path and open it
            for uri in uris_list:
                path = uri.path()
                if sys.platform.startswith('win'):
                    path = path[1:]
                if os.path.isfile(path):
                    self.vtapp.fileOpen(path)
            return True

        # If the dropped object is a tree node then update the tree
        parent_node = self.nodeFromIndex(parent)
        encoded_data = data.data("application/x-dbstreemodeldatalist")

        # Moving is not allowed if the parent group remains the same
        initial_parent = self.initial_parent
        if parent == initial_parent:
            return False

        stream = QtCore.QDataStream(encoded_data, QtCore.QIODevice.ReadOnly)
        while not stream.atEnd():
            # Decode the encoded data
            filepath = stream.readQString()
            nodepath = stream.readQString()
            initial_row = int(stream.readQString())

            # A node cannot be moved on itself
            if (parent_node.filepath, parent_node.nodepath) == (filepath,
                                                                nodepath):
                return False

            # Move the node to its final destination in the PyTables database
            new_name = self.move_node(filepath, nodepath, parent)
            if new_name is None:
                return False

            # Remove the dragged node from the model
            self.remove_rows(initial_row, parent=initial_parent)

            # If needed, refresh the copied node info
            try:
                if self.ccni['filepath'] == filepath:
                    if self.ccni['nodepath'].startswith(nodepath):
                        self.ccni = {}
            except KeyError:
                pass

            # Add the dropped node to the model
            self.lazyAddChildren(parent)
            parent_node.updated = True

            # Select the pasted node
            self.selectIndex(parent, new_name)

        return True



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#       Copyright (C) 2005-2007 Carabos Coop. V. All rights reserved
#       Copyright (C) 2008-2019 Vicent Mas. All rights reserved
#
#       This program is free software: you can redistribute it and/or modify
#       it under the terms of the GNU General Public License as published by
#       the Free Software Foundation, either version 3 of the License, or
#       (at your option) any later version.
#
#       This program is distributed in the hope that it will be useful,
#       but WITHOUT ANY WARRANTY; without even the implied warranty of
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#       GNU General Public License for more details.
#
#       You should have received a copy of the GNU General Public License
#       along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#       Author:  Vicent Mas - vmas@vitables.org

"""
Here is defined the NodeItemDelegate class.
"""

__docformat__ = 'restructuredtext'

from qtpy import QtCore
from qtpy import QtGui
from qtpy import QtWidgets

import vitables.utils

translate = QtWidgets.QApplication.translate

class NodeItemDelegate(QtWidgets.QItemDelegate):
    """
    A custom delegate for editing items of the tree of databases model.

    The delegate is used for editing the text of items of the tree of
    databases model, i.e., the item's DisplayRole. In other words, it is
    used to edit the name of the nodes in a database object tree.
    """

    def __init__(self, vtgui, parent=None):
        """
        Creates the custom delegate.
        """

        super(NodeItemDelegate, self).__init__(parent)
        self.current_name = None
        self.vtgui = vtgui


    def setEditorData(self, editor, index):
        """
        Edits the data of an item with a given index.

        This method is automatically called when an editor is created,
        i.e., when the user double click an item while the `Shift` key is
        pressed.

        :Parameters:

        - `editor`: the editor widget
        - `index`: the index of the item being edited
        """

        node_name = index.model().data(index, QtCore.Qt.DisplayRole)
        self.current_name = node_name
        editor.setText(node_name)


    def setModelData(self, editor, model, index):
        """
        Update the model with the just edited data.

        This method is called automatically when an editor is closed,
        i.e., when the user editing ends.

        :Parameters:

        - `editor`: the editor widget
        - `model`: the model whose data is being setup
        - `index`: the index of the item being edited
        """

        suggested_nodename = editor.text()

        node = model.nodeFromIndex(index)  # A GroupNode or a LeafNode instance
        parent = node.parent

        #
        # Check if the nodename is already in use
        #
        sibling = getattr(parent.node, '_v_children').keys()
        # Note that current nodename is not allowed as new nodename.
        # Embedding it in the pattern makes unnecessary to pass it to the
        # rename dialog via method argument and simplifies the code
        pattern = """(^{0}$)|""" \
            """(^[a-zA-Z_]+[0-9a-zA-Z_ ]*)""".format(self.current_name)
        info = [translate('NodeItemDelegate',
            'Renaming a node: name already in use',
            'A dialog caption'),
            translate('NodeItemDelegate',
                """Source file: {0}\nParent group: {1}\n\nThere is """
                """already a node named '{2}' in that parent group.\n""",
                'A dialog label').format\
                    (parent.filepath, parent.nodepath, suggested_nodename)]
        # Validate the nodename
        nodename, overwrite = vitables.utils.getFinalName(suggested_nodename,
            sibling, pattern, info)
        if nodename is None:
            editor.setText(self.current_name)
            return

        # Update the underlying data structure
        model.rename_node(index, nodename, overwrite)
        self.closeEditor.emit(editor, 0)

        # Update the application status bar
        self.vtgui.updateStatusBar()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#       Copyright (C) 2005-2007 Carabos Coop. V. All rights reserved
#       Copyright (C) 2008-2019 Vicent Mas. All rights reserved
#
#       This program is free software: you can redistribute it and/or modify
#       it under the terms of the GNU General Public License as published by
#       the Free Software Foundation, either version 3 of the License, or
#       (at your option) any later version.
#
#       This program is distributed in the hope that it will be useful,
#       but WITHOUT ANY WARRANTY; without even the implied warranty of
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#       GNU General Public License for more details.
#
#       You should have received a copy of the GNU General Public License
#       along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#       Author:  Vicent Mas - vmas@vitables.org

"""
This module provides a dialog for changing ``ViTables`` settings at runtime.

The dialog has 3 pages managed via QtGui.QStackedWidget: General settings
page, Look&Feel settings page and Plugins settings page.
"""

import os

from qtpy import QtCore
from qtpy import QtGui
from qtpy import QtWidgets

from qtpy.uic import loadUiType

from vitables.vtsite import ICONDIR
import vitables.utils


__docformat__ = 'restructuredtext'

translate = QtWidgets.QApplication.translate
# This method of the PyQt5.uic module allows for dynamically loading user
# interfaces created by QtDesigner. See the PyQt5 Reference Guide for more
# info.
Ui_SettingsDialog = \
    loadUiType(os.path.join(os.path.dirname(__file__), 'settings_dlg.ui'))[0]


class Preferences(QtWidgets.QDialog, Ui_SettingsDialog):
    """
    Create the Settings dialog.

    By loading UI files at runtime we can:

        - create user interfaces at runtime (without using pyuic)
        - use multiple inheritance, MyParentClass(BaseClass, FormClass)

    """

    def __init__(self):
        """
        Initialize the preferences dialog.

        * initializes the dialog appearance according to current preferences
        * connects dialog widgets to slots that provide them functionality
        """

        self.vtapp = vitables.utils.getVTApp()
        self.vtgui = self.vtapp.gui
        # Create the Settings dialog and customize it
        super(Preferences, self).__init__(self.vtgui)
        self.setupUi(self)

        self.config = self.vtapp.config
        self.pg_loader = self.vtapp.plugins_mgr
        self.all_plugins = \
            dict(item for item in self.pg_loader.all_plugins.items())
        self.enabled_plugins = self.pg_loader.enabled_plugins[:]

        # Setup the Plugins page
        self.setupPluginsPage()

        # Setup the page selector widget
        self.setupSelector()

        # Display the General Settings page
        self.stackedPages.setCurrentIndex(0)

        # Style names can be retrieved with qt.QStyleFactory.keys()
        styles = QtWidgets.QStyleFactory.keys()
        self.stylesCB.insertItems(0, styles)

        # The dictionary of current ViTables preferences
        self.initial_prefs = {}
        style_sheet = self.vtgui.logger.styleSheet()
        paper = style_sheet[-7:]
        self.initial_prefs['Logger/Paper'] = QtGui.QColor(paper)
        self.initial_prefs['Logger/Text'] = self.vtgui.logger.textColor()
        self.initial_prefs['Logger/Font'] = self.vtgui.logger.font()
        self.initial_prefs['Workspace/Background'] = \
            self.vtgui.workspace.background()
        self.initial_prefs['Look/currentStyle'] = self.config.current_style
        self.initial_prefs['Session/startupWorkingDir'] = \
            self.config.initial_working_directory
        self.initial_prefs['Session/restoreLastSession'] = \
            self.config.restore_last_session

        # The dictionary used to update the preferences
        self.new_prefs = {}

        # Apply the current ViTables configuration to the Preferences dialog
        self.resetPreferences()

        # Connect SIGNALS to SLOTS
        self.buttonsBox.helpRequested.connect(
            QtWidgets.QWhatsThis.enterWhatsThisMode)

    def setupPluginsPage(self):
        """Populate the tree of plugins.
        """

        nrows = len(self.all_plugins)
        self.plugins_model = QtGui.QStandardItemModel(nrows, 2, self)
        self.pluginsTV.setModel(self.plugins_model)
        header = QtWidgets.QHeaderView(QtCore.Qt.Horizontal, self.pluginsTV)
        header.setStretchLastSection(True)
        self.pluginsTV.setHeader(header)
        self.plugins_model.setHorizontalHeaderLabels(['Name', 'Comment'])

        # Populate the model
        row = 0
        for UID, desc in self.all_plugins.items():
            name = desc['name']
            comment = desc['comment']
            nitem = QtGui.QStandardItem(name)
            nitem.setData(UID)
            nitem.setCheckable(True)
            if UID in self.enabled_plugins:
                nitem.setCheckState(QtCore.Qt.Checked)
            citem = QtGui.QStandardItem(comment)
            self.plugins_model.setItem(row, 0, nitem)
            self.plugins_model.setItem(row, 1, citem)
            row = row + 1

    def setupSelector(self):
        """Setup the page selector widget of the Preferences dialog.
        """

        iconsdir = os.path.join(ICONDIR, '64x64')
        self.selector_model = QtGui.QStandardItemModel(self)
        self.pageSelector.setModel(self.selector_model)

        # Populate the model with top level items
        alignment = QtCore.Qt.AlignLeft | QtCore.Qt.AlignVCenter
        flags = QtCore.Qt.ItemIsSelectable | QtCore.Qt.ItemIsEnabled
        general_item = QtGui.QStandardItem()
        general_item.setIcon(QtGui.QIcon(os.path.join(
            iconsdir, 'preferences-other.png')))
        general_item.setText(translate('Preferences', "  General  ",
                                       "Text for page selector icon"))
        general_item.setTextAlignment(alignment)
        general_item.setFlags(flags)

        style_item = QtGui.QStandardItem()
        style_item.setIcon(QtGui.QIcon(os.path.join(
            iconsdir, 'preferences-desktop-theme.png')))
        style_item.setText(translate('Preferences', "Look & Feel",
                                     "Text for page selector icon"))
        style_item.setTextAlignment(alignment)
        style_item.setFlags(flags)

        self.plugins_item = QtGui.QStandardItem()
        self.plugins_item.setIcon(QtGui.QIcon(os.path.join(
            iconsdir, 'preferences-plugin.png')))
        self.plugins_item.setText(translate('Preferences', "  Plugins  ",
                                            "Text for page selector icon"))
        self.plugins_item.setTextAlignment(alignment)
        self.plugins_item.setFlags(flags)

        for item in (general_item, style_item, self.plugins_item):
            self.selector_model.appendRow(item)

        # Add items for *loaded* plugins to the Plugins item
        index = self.selector_model.indexFromItem(self.plugins_item)
        self.pageSelector.setExpanded(index, True)
        for UID in self.vtapp.plugins_mgr.loaded_plugins.keys():
            name = UID.split('#@#')[0]
            item = QtGui.QStandardItem(name)
            item.setData(UID)
            self.plugins_item.appendRow(item)

    @QtCore.Slot("QModelIndex", name="on_pageSelector_clicked")
    def changeSettingsPage(self, index):
        """Slot for changing the selected page in the Settings dialog.

        :Parameter index: the index clicked by the user
        """

        # If top level item is clicked
        if not index.parent().isValid():
            self.stackedPages.setCurrentIndex(index.row())
        # If a plugin item is clicked
        elif index.parent() == self.plugins_item.index():
            pluginID = self.selector_model.itemFromIndex(index).data()
            self.aboutPluginPage(pluginID)

    @QtCore.Slot("QAbstractButton *", name="on_buttonsBox_clicked")
    def executeButtonAction(self, button):
        """Slot that manages button box clicks in the Preferences dialog.

        Whenever one of the `Help`, `Reset`, `Cancel` or `OK` buttons is
        clicked in the Preferences dialog this slot is called.

        :Parameter button: the clicked button.
        """

        if button == self.buttonsBox.button(QtWidgets.QDialogButtonBox.Reset):
            self.resetPreferences()
        elif button == self.buttonsBox.button(QtWidgets.QDialogButtonBox.Help):
            pass
        elif button == self.buttonsBox.button(
                QtWidgets.QDialogButtonBox.Cancel):
            self.reject()
        else:
            self.applySettings()

    def resetPreferences(self):
        """
        Apply the current ``ViTables`` configuration to the Preferences dialog.
        """

        # Startup page
        if self.initial_prefs['Session/startupWorkingDir'] == 'last':
            self.lastDirCB.setChecked(True)
        else:
            self.lastDirCB.setChecked(False)

        self.restoreCB.setChecked(
            self.initial_prefs['Session/restoreLastSession'])

        # Style page
        self.sampleTE.selectAll()
        self.sampleTE.setCurrentFont(self.initial_prefs['Logger/Font'])
        self.sampleTE.setTextColor(self.initial_prefs['Logger/Text'])
        self.sampleTE.moveCursor(QtGui.QTextCursor.End)  # Unselect text
        self.sampleTE.setStyleSheet("background-color: {0}".format(
            self.initial_prefs['Logger/Paper'].name()))

        self.workspaceLabel.setStyleSheet('background-color: {0}'.format(
            self.initial_prefs['Workspace/Background'].color().name()))

        index = self.stylesCB.findText(self.initial_prefs['Look/currentStyle'])
        self.stylesCB.setCurrentIndex(index)

        # The visual update done above is not enough, we must reset the
        # new preferences dictionary and the list of enabled plugins
        self.new_prefs.clear()
        self.new_prefs.update(self.initial_prefs)
        self.enabled_plugins = self.pg_loader.enabled_plugins[:]
        self.all_plugins = \
            dict(item for item in self.pg_loader.all_plugins.items())
#        UIDs = self.all_plugins.keys()
        for row in range(0, self.plugins_model.rowCount()):
            item = self.plugins_model.item(row, 0)
            if item.data() in self.enabled_plugins:
                item.setCheckState(2)
            else:
                item.setCheckState(0)

    def applySettings(self):
        """
        Apply the current preferences to the application and close the dialog.

        This method is a slot connected to the `accepted` signal. See
        ctor for details.
        """

        # Update the plugins manager
        self.updatePluginsManager()

        # Update the rest of settings
        for key, value in self.new_prefs.items():
            self.new_prefs[key] = value

        self.accept()

    @QtCore.Slot("bool", name="on_lastDirCB_toggled")
    def setInitialWorkingDirectory(self, cb_on):
        """
        Configure startup behavior of the application.

        If the `Start in last opened directory` check box is checked
        then when the user opens a file *for the very first time* the
        current directory of the file selector dialog (CDFSD) will be
        the last directory accessed in the previous ``ViTables session``. If
        it is not checked then ``ViTables`` follows the standard behavior:
        if it has been started from a console session then the CDFSD
        will be the current working directory of the session, if it has
        been started from a menu/desktop-icon/run-command-applet the
        CDFSD will be the users' home.

        This is a slot method.

        :Parameter cb_on: a boolean indicator of the checkbox state.
        """

        if cb_on:
            self.new_prefs['Session/startupWorkingDir'] = 'last'
        else:
            self.new_prefs['Session/startupWorkingDir'] = 'home'

    @QtCore.Slot("bool", name="on_restoreCB_toggled")
    def setRestoreSession(self, cb_on):
        """
        Configure startup behavior of the application.

        If the `Restore last session` checkbox is checked then, at the
        next startup, the application will atempt to restore the last
        working session.

        This is a slot method.

        :Parameter cb_on: a boolean indicator of the checkbox state.
        """

        if cb_on:
            self.new_prefs['Session/restoreLastSession'] = True
        else:
            self.new_prefs['Session/restoreLastSession'] = False

    @QtCore.Slot(name="on_fontPB_clicked")
    def setLoggerFont(self):
        """Slot for setting the logger font."""

        new_font, is_ok = \
            QtWidgets.QFontDialog.getFont(self.sampleTE.currentFont())
        # The selected font is applied to the sample text
        if is_ok:
            self.new_prefs['Logger/Font'] = new_font
            self.sampleTE.selectAll()
            self.sampleTE.setCurrentFont(new_font)
            self.sampleTE.moveCursor(QtGui.QTextCursor.End)  # Unselect text

    @QtCore.Slot(name="on_foregroundPB_clicked")
    def setLoggerTextColor(self):
        """Slot for setting the logger foreground color."""

        text_color = self.sampleTE.textColor()
        color = QtWidgets.QColorDialog.getColor(text_color)
        # The selected text color is applied to the sample text
        if color.isValid():
            self.new_prefs['Logger/Text'] = color
            self.sampleTE.selectAll()
            self.sampleTE.setTextColor(color)
            self.sampleTE.moveCursor(QtGui.QTextCursor.End)

    @QtCore.Slot(name="on_backgroundPB_clicked")
    def setLoggerBackgroundColor(self):
        """Slot for setting the logger background color."""

        stylesheet = self.sampleTE.styleSheet()
        background = stylesheet[-7:]
        color = QtWidgets.QColorDialog.getColor(QtGui.QColor(background))
        # The selected paper color is applied to the sample text window
        if color.isValid():
            self.new_prefs['Logger/Paper'] = color
            new_stylesheet = stylesheet.replace(background, color.name())
            self.sampleTE.setStyleSheet(new_stylesheet)

    @QtCore.Slot(name="on_workspacePB_clicked")
    def setWorkspaceColor(self):
        """Slot for setting the workspace background color."""

        stylesheet = self.workspaceLabel.styleSheet()
        background = stylesheet[-7:]
        color = QtWidgets.QColorDialog.getColor(QtGui.QColor(background))
        # The selected color is applied to the sample label besides the button
        if color.isValid():
            self.new_prefs['Workspace/Background'] = QtGui.QBrush(color)
            new_stylesheet = stylesheet.replace(background, color.name())
            self.workspaceLabel.setStyleSheet(new_stylesheet)

    @QtCore.Slot("QString", name="on_stylesCB_activated")
    def setGlobalStyle(self, style_name):
        """
        Slot for setting the application style.

        :Parameter style_name: the style to be applied
        """
        self.new_prefs['Look/currentStyle'] = style_name

    def updatePluginsManager(self):
        """Update the plugins manager before closing the dialog.

        When the Apply button is clicked the list of enabled plugins
        is refreshed.
        """

        self.enabled_plugins = []
        for row in range(self.plugins_model.rowCount()):
            item = self.plugins_model.item(row, 0)
            if item.checkState() == 2:
                self.enabled_plugins.append(item.data())

        self.pg_loader.enabled_plugins = self.enabled_plugins[:]

    def aboutPluginPage(self, pluginID):
        """A page with info about the plugin clicked in the selector widget.

        :Parameter pluginID: a unique ID for getting the proper plugin
        """

        # Refresh the Preferences dialog pages. There is at most one
        # About Plugin page at any given time
        while self.stackedPages.count() > 3:
            about_page = self.stackedPages.widget(3)
            self.stackedPages.removeWidget(about_page)
            del about_page

        pg_instance = self.vtapp.plugins_mgr.loaded_plugins[pluginID]
        try:
            about_page = pg_instance.helpAbout(self.stackedPages)
        except AttributeError:
            about_page = QtWidgets.QWidget(self.stackedPages)
            label = QtWidgets.QLabel(translate(
                'Preferences',
                'Sorry, there are no info available for this plugin',
                'A text label'), about_page)
            layout = QtWidgets.QVBoxLayout(about_page)
            layout.addWidget(label)

        self.stackedPages.addWidget(about_page)
        self.stackedPages.setCurrentIndex(3)



#       Copyright (C) 2008-2019 Vicent Mas. All rights reserved
#
#       This program is free software: you can redistribute it and/or modify
#       it under the terms of the GNU General Public License as published by
#       the Free Software Foundation, either version 3 of the License, or
#       (at your option) any later version.
#
#       This program is distributed in the hope that it will be useful,
#       but WITHOUT ANY WARRANTY; without even the implied warranty of
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#       GNU General Public License for more details.
#
#       You should have received a copy of the GNU General Public License
#       along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#       Author:  Vicent Mas - vmas@vitables.org

"""Plugin that provides export of `tables.Leaf` nodes into `CSV` files.

When exporting tables, a header with the field names can be inserted.

In general, tables/arrays with Ndimensional fields are not exported because they
are not written by np.savetxt() in a way compliant with the CSV format in which
each line of the file is a data record.

Neither numpy scalar arrays are exported.
"""


import logging
import os

import numpy
import tables
from qtpy import QtCore
from qtpy import QtGui
from qtpy import QtWidgets

import vitables.csv.csvutils as csvutils
import vitables.utils

__docformat__ = 'restructuredtext'

translate = QtWidgets.QApplication.translate

log = logging.getLogger(__name__)


class ExportToCSV(QtCore.QObject):
    """Provides `CSV` export capabilities for arrays.
    """

    def __init__(self):
        """The class constructor.
        """

        super(ExportToCSV, self).__init__()

        # Get a reference to the application instance
        self.vtapp = vitables.utils.getVTApp()
        if self.vtapp is None:
            return

        self.vtgui = vitables.utils.getGui()

        # Add an entry under the Dataset menu
        self.icons_dictionary = vitables.utils.getIcons()
        self.addEntry()

        # Connect signals to slots
        self.vtgui.dataset_menu.aboutToShow.connect(self.updateDatasetMenu)
        self.vtgui.leaf_node_cm.aboutToShow.connect(self.updateDatasetMenu)

    def addEntry(self):
        """Add the `Export to CSV..`. entry to `Dataset` menu.
        """

        self.export_csv_action = QtWidgets.QAction(
            translate('ExportToCSV', "E&xport to CSV...",
                      "Save dataset as CSV"),
            self,
            shortcut=QtGui.QKeySequence.UnknownKey, triggered=self.export,
            icon=vitables.utils.getIcons()['document-export'],
            statusTip=translate(
                'ExportToCSV',
                "Save the dataset as a plain text with CSV format",
                "Status bar text for the Dataset -> Export to CSV... action"))
        self.export_csv_action.setObjectName('export_csv')

        # Add the action to the Dataset menu
        vitables.utils.addToMenu(self.vtgui.dataset_menu, self.export_csv_action)

        # Add the action to the leaf context menu
        vitables.utils.addToLeafContextMenu(self.export_csv_action)

    def updateDatasetMenu(self):
        """Update the `export` QAction when the Dataset menu is pulled down.

        This method is a slot. See class ctor for details.
        """

        enabled = True
        current = self.vtgui.dbs_tree_view.currentIndex()
        if current:
            leaf = self.vtgui.dbs_tree_model.nodeFromIndex(current)
            if leaf.node_kind in ('group', 'root group'):
                enabled = False

        self.export_csv_action.setEnabled(enabled)

    def getExportInfo(self, is_table):
        """Get info about the file where dataset will be stored.

        The info is retrieved from the FileSelector dialog. The returned info
        is the filepath and whether or not a header must be added.

        :Parameter is_table: True if the exported dataset is a tables.Table
          instance
        """

        # Call the file selector (and, if needed, customise it)
        file_selector = vitables.utils.getFileSelector(
            self.vtgui,
            translate('ExportToCSV', 'Exporting dataset to CSV format',
                      'Caption of the Export to CSV dialog'),
            dfilter=translate('ExportToCSV', """CSV Files (*.csv);;"""
                              """All Files (*)""",
                              'Filter for the Export to CSV dialog'),
            settings={'accept_mode': QtWidgets.QFileDialog.AcceptSave,
                      'file_mode': QtWidgets.QFileDialog.AnyFile,
                      'history': self.vtapp.file_selector_history,
                      'label': translate('ExportToCSV', 'Export',
                                         'Accept button text for QFileDialog')}
        )

        # Customise the file selector dialog for exporting to CSV files
        if is_table:
            # We can get the layout of Qt dialogs but not of native dialogs
            file_selector.setOption(QtWidgets.QFileDialog.DontUseNativeDialog,
                                    True)
            fs_layout = file_selector.layout()
            header_label = QtWidgets.QLabel('Add header:', file_selector)
            header_cb = QtWidgets.QCheckBox(file_selector)
            header_cb.setChecked(True)
            fs_layout.addWidget(header_label, 4, 0)
            fs_layout.addWidget(header_cb, 4, 1)

        # Execute the dialog
        try:
            if file_selector.exec_():  # OK clicked
                filepath = file_selector.selectedFiles()[0]
                # Make sure filepath contains no backslashes
                filepath = QtCore.QDir.fromNativeSeparators(filepath)
                filepath = csvutils.checkFilenameExtension(filepath)
                # Update the working directory
                working_dir = file_selector.directory().canonicalPath()
            else:  # Cancel clicked
                filepath = working_dir = ''
        finally:
            add_header = False
            if is_table:
                add_header = header_cb.isChecked()
            del file_selector

        # Process the returned values
        if not filepath:
            # The user has canceled the dialog
            return

        # Update the history of the file selector widget
        self.vtapp.updateFSHistory(working_dir)

        # Check the returned path
        if os.path.exists(filepath):
            log.error(translate(
                'ExportToCSV',
                'Export failed because destination file already exists.',
                'A file creation error'))
            return

        if os.path.isdir(filepath):
            log.error(translate(
                'ExportToCSV',
                'Export failed because destination container is a directory.',
                'A file creation error'))
            return

        return filepath, add_header

    # def _try_exporting_dataframe(self, leaf):
    #     ## FIXME: Hack to export to csv.
    #     #
    #     from ...vttables import df_model
    #
    #     leaf_model = df_model.try_opening_as_dataframe(leaf)
    #     if not leaf_model:
    #         return
    #
    #     export_info = self.getExportInfo(is_table=True)
    #     if export_info is None:
    #         return
    #
    #     leaf_model.to_csv(*export_info)
    #     return True

    def export(self):
        """Export a given dataset to a `CSV` file.

        This method is a slot connected to the `export` QAction. See the
        :meth:`addEntry` method for details.
        """

        # The PyTables node tied to the current leaf of the databases tree
        current = self.vtgui.dbs_tree_view.currentIndex()
        leaf = self.vtgui.dbs_tree_model.nodeFromIndex(current).node

        # Empty datasets aren't saved as CSV files
        if leaf.nrows == 0:
            log.info(translate(
                'ExportToCSV', 'Empty dataset. Nothing to export.'))
            return

        # Scalar arrays aren't saved as CSV files
        if leaf.shape == ():
            log.info(translate(
                'ExportToCSV', 'Scalar array. Nothing to export.'))
            return

        # Datasets with more than 3 dimensions aren't saved as CSV files
        # (see module's docstring)
        if len(leaf.shape) > 3:
            log.info(translate(
                'ExportToCSV', 'The selected node has more than '
                '3 dimensions. I can\'t export it to CSV format.'))
            return

        # Variable lenght arrays aren't saved as CSV files
        if isinstance(leaf, tables.VLArray):
            log.info(translate(
                'ExportToCSV', 'The selected node is a VLArray. '
                'I can\'t export it to CSV format.'))
            return

        # Tables with Ndimensional fields aren't saved as CSV files
        is_table = isinstance(leaf, tables.Table)
        if is_table:
            if self._try_exporting_dataframe(leaf):
                return

            first_row = leaf[0]
            for item in first_row:
                if item.shape != ():
                    log.info(translate(
                        'ExportToCSV',
                        'Some fields aren\'t scalars. '
                        'I can\'t export the table to CSV format.'))
                    return

        # Get the required info for exporting the dataset
        export_info = self.getExportInfo(is_table)
        if export_info is None:
            return
        else:
            filepath, add_header = export_info

        try:
            QtWidgets.qApp.setOverrideCursor(QtCore.Qt.WaitCursor)
            with open(filepath, 'ab') as out_handler:
                if add_header:
                    from functools import reduce
                    header = reduce(lambda x, y: '{0}, {1}'.format(x, y),
                                    leaf.colnames)
                    # To be consistent with numpy.savetxt use \n line breaks
                    out_handler.write(bytearray(header + '\n', 'UTF-8'))
                chunk_size = 10000
                nrows = leaf.nrows
                if chunk_size > nrows:
                    chunk_size = nrows
                # Behavior of np.divide in Python 2 and Python 3 is different so
                # we must explicitly ensure we get an integer
                nchunks = numpy.floor_divide(nrows, chunk_size)
                for i in numpy.arange(0, nchunks + 1):
                    QtWidgets.qApp.processEvents()
                    cstart = chunk_size * i
                    if cstart >= nrows:
                        break
                    cstop = cstart + chunk_size
                    if cstop > nrows:
                        cstop = nrows
                    numpy.savetxt(out_handler, leaf.read(cstart, cstop, 1),
                                  fmt='%s', delimiter=',')
        except OSError:
            vitables.utils.formatExceptionInfo()
        finally:
            QtWidgets.qApp.restoreOverrideCursor()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#       Copyright (C) 2005-2007 Carabos Coop. V. All rights reserved
#       Copyright (C) 2008-2019 Vicent Mas. All rights reserved
#
#       This program is free software: you can redistribute it and/or modify
#       it under the terms of the GNU General Public License as published by
#       the Free Software Foundation, either version 3 of the License, or
#       (at your option) any later version.
#
#       This program is distributed in the hope that it will be useful,
#       but WITHOUT ANY WARRANTY; without even the implied warranty of
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#       GNU General Public License for more details.
#
#       You should have received a copy of the GNU General Public License
#       along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#       Author:  Vicent Mas - vmas@vitables.org
"""
This module implements a model (in the `MVC` sense) for the real data stored
in a Pandas HDFStore.
"""

__docformat__ = 'restructuredtext'

import logging

from qtpy import QtCore, QtGui
from qtpy.QtCore import Qt

from . import leaf_model

log = logging.getLogger(__name__)

_axis_font = QtGui.QFont()
_axis_font.setBold(True)

_axis_label_font = QtGui.QFont()
_axis_label_font.setBold(True)
_axis_label_font.setItalic(True)


def try_opening_as_dataframe(leaf):
    try:
        from pandas.io import pytables
    except ImportError:
        return

    class HDFStoreWrapper(pytables.HDFStore):
        """Subclassed to construct stores from :class:`h5py.File` instances."""

        def __init__(self, tables_h5file):
            self._path = tables_h5file.filename
            self._mode = tables_h5file.mode
            self._handle = tables_h5file
            self._filters = tables_h5file.filters

            pytables._tables()

    pgroup = leaf._g_getparent()
    try:
        assert pgroup._c_classid == 'GROUP'
    except AssertionError:
        log('Parent of {0} is not a GROUP'.format(leaf))

    pandas_attr = getattr(pgroup._v_attrs, 'pandas_type', None)
    if pandas_attr in ['frame', 'frame_table']:
        hstore = HDFStoreWrapper(leaf._v_file)

        return DataFrameModel(leaf, hstore)


def get_index_name(index, i, fallback_pattern):
    """Returns the name of a dataframe index."""
    val = i
    try:
        val = index.names[i]
    except:  # @IgnorePep8
        pass

    if val is None:
        val = fallback_pattern % i
    elif isinstance(val, int):
        val = fallback_pattern % val

    return val


class DataFrameModel(QtCore.QAbstractTableModel):
    """
    The model for data contained in pandas DataFrame chunks.

    :param parent:
        The parent of the model, passed as is in the superclass.
    :attribute _leaf:
        the underlying hdf5 data
    :attribute leaf_numrows:
        the total number of rows in the underlying data
    :attribute numrows:
        The number of rows visible which equals the chunking-size.
    :attribute numcols:
        The total number of columns visible, equal to those visible.
    :attribute start:
        The zero-based starting index of the chunk within the total rows.
    :attribute _nheaders:
        A tuple ``(row_span, col_span)`` for the number of *columns/indices*
        headers respectively, in case they are pandas multi-index, or
        just ``(1, 1)``.
    """

    def __init__(self, leaf, hstore, parent=None):
        """Create the model.
        """
        self._leaf = leaf
        self._pgroup = self._leaf._g_getparent()._v_pathname

        self._hstore = hstore
        self.start = 0

        # The dataset number of rows is potentially huge but tables are
        #  kept small: just the data returned by a read operation of the
        #  buffer are displayed
        self.leaf_numrows = leaf.shape[0]
        self.numrows = min(self.leaf_numrows, leaf_model.CHUNK_SIZE)

        # Track selected cell.
        self.selected_cell = {'index': QtCore.QModelIndex(), 'buffer_start': 0}

        # Populate the model with the first chunk of data.
        self.loadData(0, self.numrows)
        chunk = self._chunk
        self.numcols = len(chunk.columns)

        def count_multiindex(index):
            try:
                return index.nlevels
            except AttributeError:
                return 1

        # Number of levels in the index/columns indexes
        self._nheaders = tuple(count_multiindex(idx)
                               for idx
                               in (chunk.columns, chunk.index))

        super(DataFrameModel, self).__init__(parent)

    def columnCount(self, index=QtCore.QModelIndex()):
        """
        The total number of columns, or number of children for the given index.

        :param index:
            Overrides should return 0 for valid indices (if no children exist)
            or the number of the total *columns* (including row labels) exposed
            by the model.
        """
        return 0 if index.isValid() else self.numcols + self._nheaders[1]

    def rowCount(self, index=QtCore.QModelIndex()):
        """
        The total number of rows, or number of children for the given index.

        :param index:
            Overrides should return 0 for valid indices (if no children exist)
            or the number of the total *rows* (including column labels) exposed
            by the model.
        """

        return 0 if index.isValid() else self.numrows + self._nheaders[0]

    def loadData(self, start, length):
        """Load the model with fresh chunk from the underlying leaf.

        :param start:
            The first row (within the total nrows) of the chunk to read.
        :param length:
            The buffer size, i.e. the number of rows to be read.
        """
        # Enforce scroll limits.
        #
        start = max(start, 0)
        stop = min(start + length, self.leaf_numrows)
        try:
            assert stop >= start
        except AssertionError:
            log(('Chunk: number of rows {0}, first row {1}, '
                 'last row {2}').format(self.numrows, start, stop))

        # Ensure that the whole buffer will be filled.
        #
        actual_start = stop - self.numrows
        start = max(min(actual_start, start), 0)

        self._chunk = self._hstore.select(self._pgroup,
                                          start=start, stop=stop)
        self.start = start

    def get_corner_span(self):
        """Must return ``(row_span, col_span)`` tuple for the top-left cell."""
        return self._nheaders if self.start == 0 else (1, 1)

    def headerData(self, section, orientation, role):
        """
        Model method to return header content and formatting.

        :param section:
            The zero-based row/column number to return.
        :param orientation:
            The header orientation (horizontal := columns, vertical := index).
        :param role:
            The Qt.XXXRole: being inspected.
        """
        if role == Qt.TextAlignmentRole:
            if orientation == Qt.Horizontal:
                return Qt.AlignCenter | Qt.AlignBottom
            return Qt.AlignRight | Qt.AlignVCenter

        n_columns, n_index = self._nheaders
        is_header = ((orientation == Qt.Horizontal
                      and section < n_index)
                     or (orientation == Qt.Vertical
                         and (self.start + section) < n_columns))

        if role == Qt.DisplayRole:
            if orientation == Qt.Horizontal:
                if is_header:
                    return get_index_name(self._chunk.index, section, 'I%s')
                return str(section)

            if orientation == Qt.Vertical:
                if is_header:
                    return get_index_name(self._chunk.columns, section, 'C%s')
                return str(self.start + section)

        if role == Qt.FontRole and is_header:
            return _axis_label_font

    def data(self, index, role=Qt.DisplayRole):
        """Returns the data stored under the given role for the item
        referred to by the index.

        This is an overwritten method.

        :Parameters:

        - `index`: the index of a data item
        - `role`: the role being returned
        """
        row, col = index.row(), index.column()
        n_columns, n_index = self._nheaders
        df = self._chunk

        if not index.isValid() or not (0 <= row < (self.numrows + n_columns)):
            return None

        is_index = col < n_index
        is_columns = (self.start + row) < n_columns

        if is_index and is_columns:
            return None

        if is_index:
            if role == Qt.DisplayRole:
                val = df.index[row - n_columns]
                if n_index > 1:
                    val = val[col]
                return str(val)
            if role == Qt.FontRole:
                return _axis_font
            if role == Qt.TextAlignmentRole:
                return int(Qt.AlignRight | Qt.AlignVCenter)
            return

        if is_columns:
            if role == Qt.DisplayRole:
                val = df.columns[col - n_index]
                if n_columns > 1:
                    val = val[row]
                return str(val)
            if role == Qt.FontRole:
                return _axis_font
            if role == Qt.TextAlignmentRole:
                return int(Qt.AlignCenter | Qt.AlignBottom)
            return

        if role == Qt.DisplayRole:
            return str(self._chunk.iat[row - n_columns, col - n_index])

        # if role == Qt.TextAlignmentRole:
        #     return int(Qt.AlignLeft|Qt.AlignTop)

        return

    def cell(self, row, col):
        return None  # Disable zoom.

    def to_csv(self, filepath, add_header):
        import io

        with io.open(filepath, 'wt', encoding='utf-8') as fd:
            chunks = iter(self._hstore.select(self._pgroup, iterator=True))
            df = next(chunks)
            df.to_csv(fd, header=add_header)

            for df in chunks:
                df.to_csv(fd, header=False)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#       Copyright (C) 2005-2007 Carabos Coop. V. All rights reserved
#       Copyright (C) 2008-2019 Vicent Mas. All rights reserved
#
#       This program is free software: you can redistribute it and/or modify
#       it under the terms of the GNU General Public License as published by
#       the Free Software Foundation, either version 3 of the License, or
#       (at your option) any later version.
#
#       This program is distributed in the hope that it will be useful,
#       but WITHOUT ANY WARRANTY; without even the implied warranty of
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#       GNU General Public License for more details.
#
#       You should have received a copy of the GNU General Public License
#       along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#       Author:  Vicent Mas - vmas@vitables.org

"""
This module collects information about a given node of a `PyTables` database.
"""

__docformat__ = 'restructuredtext'

import os.path

import vitables.utils

class NodeInfo(object):
    """Collects information about a given node.

    The following data and metadata can be collected:

    * format of the database (generic `HDF5` or `PyTables`)
    * filename of the database
    * filepath of the database
    * opening mode of the database
    * size of the database
    * node type
    * node name
    * node path
    * for `tables.Group` nodes:

        - attributes set instance and related info
        - dictionary of nodes hanging from a group
        - dictionary of groups hanging from a group
        - dictionary of leaves hanging from a group

    * for `tables.Table` nodes:

        - columns names, datatypes and shapes
        - number of rows and columns
        - filters
        - shape
        - flavor

    * for `tables.XArray` nodes:

        - number of rows
        - datatype
        - filters
        - shape
        - flavor

    :Parameter node_item: an instance of :meth:`RootGroupNode`,
      :meth:`GroupNode` or :meth:`LeafNode`.
    """

    def __init__(self, node_item):
        """Collects information about a given node.

        Some PyTables2.X string attributes are regular Python strings instead
        of unicode strings and have to be explicitely converted to unicode.
        """

        self.node = node_item.node

        # The hosting File instance, filepath, filename and opening mode
        self.filepath = str(node_item.filepath)
        self.filename = os.path.basename(self.filepath)
        self.h5file = self.node._v_file
        mode = str(self.h5file.mode)
        if mode == 'a':
            self.mode = 'append'
        elif mode == 'r':
            self.mode = 'read-only'
        else:
            self.mode = 'read-write'

        # The node type is a string with one of the following values:
        # root group, group, table, vlarray, earray, carray, array
        # or unimplemented
        self.node_type = node_item.node_kind
        self.file_type = self.format + ', ' + self.size
        self.nodename = str(node_item.name)
        self.nodepath = str(node_item.nodepath)

        # The attributes set instance
        self.asi = self.node._v_attrs
        sysattrs_names = self.asi._v_attrnamessys
        self.system_attrs = \
            dict((n, getattr(self.asi, n, None)) for n in sysattrs_names)
        userattrs_names = self.asi._v_attrnamesuser
        self.user_attrs = \
            dict((n, getattr(self.asi, n, None)) for n in userattrs_names)

    # Properties for File instances

    def _format(self):
        """The format of the hosting `tables.File` instance"""

        if self.h5file._isPTFile:
            return 'PyTables file'
        else:
            return 'Generic HDF5 file'

    format = property(fget=_format)


    def _size(self):
        """The size of the hosting `tables.File` instance"""

        n_bytes = os.path.getsize(self.filepath) *1.0
        kbytes = n_bytes/1024
        mbytes = kbytes/1024
        gbytes = mbytes/1024
        tbytes = gbytes/1024
        if kbytes < 1:
            size = '{0:.0f} bytes'.format(n_bytes)
        elif mbytes < 1:
            size = '{0:.1f} KB'.format(kbytes)
        elif gbytes < 1:
            size = '{0:.1f} MB'.format(mbytes)
        elif tbytes < 1:
            size = '{0:.1f} GB'.format(gbytes)
        else:
            size = '{0:.1f} TB'.format(tbytes)
        return size

    size = property(fget=_size)

    # Properties for Group instances

    def _hangingNodes(self):
        """The dictionary of nodes hanging from this node."""

        try:
            return self.node._v_children
        except AttributeError:
            return {}

    hanging_nodes = property(fget=_hangingNodes)


    def _hangingGroups(self):
        """The dictionary of groups hanging from this node."""

        try:
            return self.node._v_groups
        except AttributeError:
            return {}

    hanging_groups = property(fget=_hangingGroups)


    def _hangingLeaves(self):
        """The dictionary of leaves hanging from this node."""

        try:
            return self.node._v_leaves
        except AttributeError:
            return {}

    hanging_leaves = property(fget=_hangingLeaves)


    def _hangingLinks(self):
        """The dictionary of links hanging from this node."""

        try:
            return self.node._v_links
        except AttributeError:
            return {}

    hanging_links = property(fget=_hangingLinks)


    # Properties for Leaf instances

    def _type(self):
        """The `PyTables` data type of the atom for `tables.Leaf` nodes."""

        if self.node_type.count('array'):
            try:
                return str(self.node.atom.type)
            except AttributeError:
                return None
        elif self.node_type == 'table':
            return 'record'

    type = property(fget=_type)


    def _nrows(self):
        """The current number of rows in the `tables.Leaf` node."""

        try:
            return self.node.shape[0]
        except TypeError:  #  shape is None
            return 0
        except IndexError:  # numpy scalar arrays have shape = ()
            return 1

    nrows = property(fget=_nrows)


    def _shape(self):
        """The shape of data in the `tables.Leaf` node."""

        try:
            return self.node.shape
        except AttributeError:
            return None

    shape = property(fget=_shape)


    def _flavor(self):
        """The type of data object read from the `tables.Leaf` node."""

        try:
            return str(self.node.flavor)
        except AttributeError:
            return None

    flavor = property(fget=_flavor)


    def _filters(self):
        """Filters property for this `tables.Leaf` node."""

        try:
            return self.node.filters
        except AttributeError:
            return None

    filters = property(fget=_filters)

    # Properties for Table instances

    def _colNames(self):
        """The list of names of top-level columns in a `tables.Table` node."""

        try:
            return self.node.colnames
        except AttributeError:
            return []

    columns_names = property(fget=_colNames)


    def _colPathNames(self):
        """The list of paths of top-level columns in a `tables.Table` node."""

        try:
            return self.node.colpathnames
        except AttributeError:
            return []

    columns_pathnames = property(fget=_colPathNames)


    def _colTypes(self):
        """
        Mapping with `tables.Table` field names and their `PyTables` datatypes.
        """

        try:
            return self.node.coltypes
        except AttributeError:
            return {}

    columns_types = property(fget=_colTypes)


    def _colShapes(self):
        """
        Mapping with the `tables.Table` field names and their shapes.
        """

        try:
            coldescrs = self.node.coldescrs
            return dict((k, v.shape) for (k, v) in coldescrs.items())
        except AttributeError:
            return {}

    columns_shapes = property(fget=_colShapes)


    def _ncolumns(self):
        """The current number of columns in the `tables.Table` node."""

        try:
            return len(self.node.columns_names)
        except AttributeError:
            return None

    ncolumns = property(fget=_ncolumns)


    def _target(self):
        """The target of a `tables.Link` node."""

        try:
            return str(self.node.target)
        except AttributeError:
            return None

    target = property(fget=_target)


    def _link_type(self):
        """The kind of link (i.e. soft or external) of a `tables.Link` node."""

        if self.target:
            try:
                link_type = 'external'
                self.node.extfile
            except AttributeError:
                link_type = 'soft'
            return link_type
        else:
            return None

    link_type = property(fget=_link_type)





#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#       Copyright (C) 2008-2019 Vicent Mas. All rights reserved
#
#       This program is free software: you can redistribute it and/or modify
#       it under the terms of the GNU General Public License as published by
#       the Free Software Foundation, either version 3 of the License, or
#       (at your option) any later version.
#
#       This program is distributed in the hope that it will be useful,
#       but WITHOUT ANY WARRANTY; without even the implied warranty of
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#       GNU General Public License for more details.
#
#       You should have received a copy of the GNU General Public License
#       along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#       Author:  Vicent Mas - vmas@vitables.org

"""
This module provides a widget with the full description of the plugin.

The module allows for configuring the plugin too.

The widget is displayed when the plugin is selected in the Preferences
dialog selector tree.
"""

import os.path
import datetime
try:
    import configparser
except ImportError:
    import ConfigParser as configparser

from qtpy import QtGui
from qtpy import QtCore
from qtpy import QtWidgets
from qtpy.uic import loadUiType


__docformat__ = 'restructuredtext'

# This method of the PyQt5.uic module allows for dynamically loading user
# interfaces created by QtDesigner. See the PyQt5 Reference Guide for more
# info.
Ui_TimeFormatterPage = \
    loadUiType(os.path.join(os.path.dirname(__file__),
                            'timeformatter_page.ui'))[0]


class AboutPage(QtWidgets.QWidget, Ui_TimeFormatterPage):
    """
    Widget for describing and customizing the Time series plugin.

    By loading UI files at runtime we can:

        - create user interfaces at runtime (without using pyuic)
        - use multiple inheritance, MyParentClass(BaseClass, FormClass)

    This widget is inserted as a page in the stacked widget of the
    Preferences dialog when the Time series item is clicked in the
    selector tree.

    """

    def __init__(self, desc, parent=None):
        """Widget for describing and customizing the Timeseries plugin.

        :Parameters:

            - `desc`: a dictionary with the plugin description
            -`parent`: the stacked widget of the Preferences dialog
        """

        # Makes the dialog and gives it a layout
        super(AboutPage, self).__init__(parent)
        self.setupUi(self)

        self.version_le.setText(desc['version'])
        self.module_name_le.setText(desc['module_name'])
        self.folder_le.setText(desc['folder'])
        self.author_le.setText(desc['author'])
        self.desc_te.setText(desc['about_text'])

        # Configuration section of the page
        self.save_button = self.buttons_box.button(
            QtWidgets.QDialogButtonBox.Save)
        self.save_button.setText('Save format')

        # The absolute path of the INI file
        self.ini_filename = \
            os.path.join(os.path.dirname(__file__), 'time_format.ini')
        # Setup initial configuration
        self.config = configparser.ConfigParser(interpolation=None)
        def_tformat = '%c'
        try:
            self.config.read_file(open(self.ini_filename))
            self.tformat = self.config['Timeseries']['strftime']
        except (IOError, configparser.ParsingError):
            self.tformat = def_tformat

        self.tformat_editor.setText(self.tformat)
        today = datetime.datetime.today().strftime(self.tformat)
        self.today_label.setText(today)

        self.save_button.clicked.connect(self.saveFormat)

    def eventFilter(self, w, e):
        """Event filter for the dialog.
        """

        if self.tformat_editor.hasFocus():
            if e.type() == 6:
                if e.key() in (QtCore.Qt.Key_Enter, QtCore.Qt.Key_Return):
                    self.applyFormat()
                    return True
        return QtWidgets.QWidget.eventFilter(self, w, e)

    def applyFormat(self):
        """
        Check the current time format.

        The time format currently entered in the line editor is applied to the
        sample label showing the today date. If the result is that expected by
        user then she can click safely the OK button. Otherwise she should try
        another time format.
        """

        today = datetime.datetime.today().strftime(self.tformat_editor.text())
        self.today_label.clear()
        self.today_label.setText(today)

    def saveFormat(self):
        """Slot for saving the entered time format.
        """

        self.config['Timeseries']['strftime'] = self.tformat_editor.text()
        with open(self.ini_filename, 'w') as ini_file:
            self.config.write(ini_file)



#       Copyright (C) 2008-2019 Vicent Mas. All rights reserved
#
#       This program is free software: you can redistribute it and/or modify
#       it under the terms of the GNU General Public License as published by
#       the Free Software Foundation, either version 3 of the License, or
#       (at your option) any later version.
#
#       This program is distributed in the hope that it will be useful,
#       but WITHOUT ANY WARRANTY; without even the implied warranty of
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#       GNU General Public License for more details.
#
#       You should have received a copy of the GNU General Public License
#       along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#       Author:  Vicent Mas - vmas@vitables.org

#
# Plugin initial draft author: Jorge Ibanez jorge.ibannez@uam.es
#

"""Convenience functions for the import_csv.py module.
"""

__docformat__ = 'restructuredtext'

import logging
import os
import re
import tempfile
import vitables.utils

import numpy
from qtpy import QtWidgets
import tables

# https://github.com/numpy/numpy/issues/10990
import warnings
warnings.filterwarnings("ignore", category=numpy.VisibleDeprecationWarning)


translate = QtWidgets.QApplication.translate
TYPE_ERROR = translate(
    'ImportCSV', 'Please, make sure that you are importing a '
    'homogeneous dataset.', 'CSV file not imported error')

log = logging.getLogger(__name__)


def getArray(buf):
    """Fill an intermediate ``numpy`` array with data read from the `CSV` file.

    The lines read from the CSV file are stored in a temporary file which is
    passed to numpy.genfromtxt() in order to create a numpy array.

    The dtypes of the numpy array are determined by the contents of each
    column. Multidimensional columns will have string datatype.

    Warning: the temporary file is written in binary mode so lines are stored
    as bytearrays (encoded as UTF-8). It means that strings in the numpy array
    will also be bytes with UTF-8 encoding and not Python 3 strings.

    :Parameter buf: the data buffer is a list of lines of the CSV file
    """

    with tempfile.TemporaryFile(mode='w+b') as temp_file:
        for line in buf:
            temp_file.write(bytearray(line, 'UTF-8'))
        temp_file.seek(0)
        data = numpy.genfromtxt(temp_file, delimiter=',', dtype=None)
    return data


def tableInfo(input_handler):
    """Return useful information about the `tables.Table` being created.

    :Parameter input_handler: the file handler of the inspected CSV file
    """

    # Inspect the CSV file reading its second line
    # (reading the first line is not safe as it could be a header)
    input_handler.seek(0)
    first_line = getArray(input_handler.readline())
    try:
        second_line = getArray(input_handler.readline())
    except IOError:
        # The second line cannot be read. We assume there is only on line
        second_line = first_line

    # Estimate the number of rows of the CSV file
    filesize = os.path.getsize(input_handler.name)
    # Record size = number of elements * element size
    record_size = second_line.size * second_line.itemsize
    nrows = filesize / record_size

    if second_line.dtype.fields is None:
        # second_line is a homogeneous array
        descr, has_header = \
            homogeneousTableInfo(input_handler, first_line, second_line)
    else:
        # second_line is a heterogeneous array
        descr, has_header = \
            heterogeneousTableInfo(input_handler, first_line, second_line)

    del second_line
    return (nrows, descr, has_header)


def heterogeneousTableInfo(input_handler, first_line, second_line):
    """Return useful information about the `tables.Table` being created.

    The `data` array is heterogenous, i.e. not all fields have the same
    dtype.

    :Parameters:

    - `input_handler`: the file handler of the inspected `CSV` file
    - `first_line`: ``numpy`` array which contains the first line of the `CSV`
      file
    - `second_line`: ``numpy`` array which contains the second line of the
      `CSV` file
    """

    has_header = False
    fl_dtype = first_line.dtype
    if (fl_dtype.fields is None) and (fl_dtype.char in('S', 'U')):
        has_header = True

    # Stuff used for finding out itemsizes of string fields
    itemsizes = {}
    for field in range(0, len(second_line.dtype)):
        if second_line.dtype[field].name.startswith('str') or \
                second_line.dtype[field].name.startswith('bytes'):
            itemsizes[field] = 0

    # If a dtype is a string, find out its biggest itemsize
    if itemsizes:
        buf_size = 1024 * 1024
        input_handler.seek(0)
        if has_header:
            # Skip the header
            input_handler.readline()
        buf = input_handler.readlines(buf_size)
        while buf:
            temp_file = tempfile.TemporaryFile()
            for line in buf:
                temp_file.write(bytearray(line, 'UTF-8'))
            for field in itemsizes.keys():
                temp_file.seek(0)
                idata = numpy.genfromtxt(temp_file, delimiter=',',
                                         usecols=(field,), dtype=None)
                itemsizes[field] = max(itemsizes[field], idata.dtype.itemsize)
                del idata
            temp_file.close()
            buf = input_handler.readlines(buf_size)

    if has_header:
        descr = {}
        for i in range(0, first_line.size):
            dtype = second_line.dtype.fields['f{0}'.format(i)][0]
            descr[first_line[i].decode('UTF-8')] = tables.Col.from_dtype(dtype,
                                                                         pos=i)
        for i in itemsizes:
            descr[first_line[i].decode(
                'UTF-8')] = tables.StringCol(itemsizes[i], pos=i)
    else:
        descr = dict([(f, tables.Col.from_dtype(t[0])) for f, t in
                      second_line.dtype.fields.items()])
        for i in itemsizes:
            descr['f{0}'.format(i)] = tables.StringCol(itemsizes[i])

    return descr, has_header


def homogeneousTableInfo(input_handler, first_line, second_line):
    """Return useful information about the `tables.Table` being created.

    The `second_line` array is homegenous, i.e. all fields have the same dtype.

    :Parameters:

    - `input_handler`: the file handler of the inspected `CSV` file
    - `first_line`: a ``numpy`` array which contains the first line of the
      `CSV` file
    - `second_line`: a ``numpy`` array which contains the second line of the
      `CSV` file
    """

    # Find out if the table has a header or not.
    has_header = False
    fldn = first_line.dtype.name
    sldn = second_line.dtype.name
    if sldn.startswith('str') or sldn.startswith('bytes'):
        answer = askForHelp(first_line)
        if answer == 'Header':
            has_header = True
    elif fldn.startswith('str') or fldn.startswith('bytes'):
        has_header = True

    input_handler.seek(0)
    if has_header:
        # Skip the header
        input_handler.readline()

    # If the fields of the table are strings then find out the biggest itemsize
    if sldn.startswith('str') or sldn.startswith('bytes'):
        itemsize = 0
        buf_size = 1024 * 1024
        buf = input_handler.readlines(buf_size)
        if not buf:
            # If the CSV file contains just one line
            itemsize = first_line.dtype.itemsize
        while buf:
            idata = getArray(buf)
            itemsize = max(itemsize, idata.dtype.itemsize)
            del idata
            buf = input_handler.readlines(buf_size)

    # Iterate over the data fields and make the table description
    # If the CSV file contains just one field then first_line is a
    # scalar array and cannot be iterated so we reshape it
    if first_line.shape == ():
        first_line = first_line.reshape(1,)
    indices = list(range(0, first_line.shape[0]))

    if has_header:
        if sldn.startswith('str') or sldn.startswith('bytes'):
            descr = dict([(first_line[i].decode('UTF-8'),
                           tables.StringCol(itemsize, pos=i))
                          for i in indices])
        else:
            descr = dict([(first_line[i].decode('UTF-8'),
                           tables.Col.from_dtype(second_line.dtype, pos=i))
                          for i in indices])
    else:
        if sldn.startswith('str') or sldn.startswith('bytes'):
            descr = dict([('f{0}'.format(field), tables.StringCol(itemsize))
                          for field in indices])
        else:
            descr = dict([('f{0}'.format(field),
                           tables.Col.from_dtype(second_line.dtype))
                          for field in indices])

    return descr, has_header


def askForHelp(first_line):
    """Ask user if the first row is a header.

    :Parameter first_line: a ``numpy`` array which contains the first line of
      the `CSV` file
    """

    title = translate('ImportCSV', 'Resolving first line role',
                      'Message box title')
    text = translate('ImportCSV', 'Does the first line of the file contain '
                     'a table header or regular data?', 'Message box text')
    itext = ''
    try:
        from functools import reduce
        dtext = reduce(lambda x, y: '{0}, {1}'.format(x, y), first_line)
    except TypeError:
        # If first_line has only one field reduce raises a TypeError
        dtext = first_line.tostring()
    buttons = {
        'Header':
        (translate('ImportCSV', 'Header', 'Button text'),
         QtWidgets.QMessageBox.YesRole),
        'Data':
        (translate('ImportCSV', 'Data', 'Button text'),
         QtWidgets.QMessageBox.NoRole),
    }
    return vitables.utils.questionBox(title, text, itext, dtext, buttons)


def earrayInfo(input_handler):
    """Return useful information about the `tables.EArray` being created.

    :Parameter input_handler: the file handler of the inspected file
    """

    # Inspect the CSV file reading its first line
    # The dtypes are determined by the contents of each column
    # Multidimensional columns will have string datatype
    first_line = getArray(input_handler.readline())

    # Estimate the number of rows of the file
    filesize = os.path.getsize(input_handler.name)
    record_size = first_line.size * first_line.itemsize
    nrows = filesize / record_size

    if first_line.dtype.name.startswith('str') or \
            first_line.dtype.name.startswith('bytes'):
        # Find out the biggest itemsize
        itemsize = 0
        buf_size = 1024 * 1024
        input_handler.seek(0)
        buf = input_handler.readlines(buf_size)
        while buf:
            idata = getArray(buf)
            itemsize = max(itemsize, idata.dtype.itemsize)
            del idata
            buf = input_handler.readlines(buf_size)
        atom = tables.StringAtom(itemsize)
    else:
        # With compound dtypes this will raise a ValueError
        atom = tables.Atom.from_dtype(first_line.dtype)

    # Get the data shape
    if nrows < 2:
        # Corner case: the file only has one row
        array_shape = (0, )
    elif first_line.shape == ():
        # Corner case: the file has just one column
        array_shape = (0, )
    else:
        # General case: the file is a MxN array
        array_shape = (0, first_line.shape[0])

    del first_line
    input_handler.seek(0)
    return nrows, atom, array_shape


def carrayInfo(input_handler):
    """Return useful information about the `tables.CArray` being created.

    :Parameter input_handler: the file handler of the inspected file
    """

    # Inspect the CSV file reading its first line
    # The dtypes are determined by the contents of each column
    # Multidimensional columns will have string datatype
    input_handler.seek(0)
    first_line = getArray(input_handler.readline())

    # This counting algorithm is faster than looping over lines with
    # fh.readline and incrementing a counter at every step
    lines = 0
    itemsize = 0
    buf_size = 1024 * 1024
    input_handler.seek(0)

    if first_line.dtype.name.startswith('str') or \
            first_line.dtype.name.startswith('bytes'):
        # Count lines and find out the biggest itemsize
        buf = input_handler.readlines(buf_size)
        while buf:
            idata = getArray(buf)
            itemsize = max(itemsize, idata.dtype.itemsize)
            del idata
            lines += len(buf)
            buf = input_handler.readlines(buf_size)
    else:
        # Count lines
        buf = input_handler.readlines(buf_size)
        while buf:
            lines += len(buf)
            buf = input_handler.readlines(buf_size)

    if itemsize:
        atom = tables.StringAtom(itemsize)
    else:
        atom = tables.Atom.from_dtype(first_line.dtype)

    # Get the data shape
    if lines == 1:
        # Corner case: the file only has one row
        array_shape = first_line.shape
        lines = first_line.shape[0]
    elif first_line.shape == ():
        # Corner case: the file has just one column
        array_shape = (lines, )
    else:
        # General case: the file is a MxN array
        array_shape = (lines, first_line.shape[0])

    del first_line
    input_handler.seek(0)
    return atom, array_shape


def isValidFilepath(filepath):
    """Check the filepath of the destination file.

    :Parameter filepath: the filepath where the imported dataset will live
    """
    valid = True
    if os.path.exists(filepath):
        log.error(translate(
            'ImportCSV',
            'CSV import failed because destination file already exists.',
            'A file creation error'))
        valid = False

    elif os.path.isdir(filepath):
        log.error(translate(
            'ImportCSV',
            'CSV import failed because destination container is a directory.',
            'A file creation error'))
        valid = False

    return valid


def checkFilenameExtension(filepath):
    """
    Check the filename extension of the CSV file.

    If the filename has no extension this method adds .csv
    extension to it.

    :Parameter filepath: the full path of the file

    :Returns: the filepath with the proper extension
    """

    if not re.search(r'\.(.+)$', os.path.basename(filepath)):
        ext = '.csv'
        filepath = filepath + ext
    return filepath



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#       Copyright (C) 2005-2007 Carabos Coop. V. All rights reserved
#       Copyright (C) 2008-2019 Vicent Mas. All rights reserved
#
#       This program is free software: you can redistribute it and/or modify
#       it under the terms of the GNU General Public License as published by
#       the Free Software Foundation, either version 3 of the License, or
#       (at your option) any later version.
#
#       This program is distributed in the hope that it will be useful,
#       but WITHOUT ANY WARRANTY; without even the implied warranty of
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#       GNU General Public License for more details.
#
#       You should have received a copy of the GNU General Public License
#       along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#       Author:  Vicent Mas - vmas@vitables.org

"""
This module provides a thin wrapper to the `tables.File` class.

The module provides methods for opening and closing PyTables databases. It also
contains methods for editing nodes (but not data stored in nodes).

"""

import logging
import os
import uuid
import vitables.utils

from qtpy import QtCore
from qtpy import QtWidgets
import tables

__docformat__ = 'restructuredtext'

translate = QtWidgets.QApplication.translate

log = logging.getLogger(__name__)


class DBDoc(QtCore.QObject):
    """
    A database contained in an hdf5/PyTables file.

    :Parameters:

    - `filePath`: full path of the DB
    - `mode`: indicate the open mode of the database file. It can be
      'r'ead-only or 'a'ppend
    - `is_tmp_dbdoc`: True if this object represents the temporary database
    """

    def __init__(self, filepath, mode, is_tmp_dbdoc=False):
        """
        Makes a data structure that defines a given database.

        A `DBDoc` instance is defined by the next attributes:

        - filepath
        - filename
        - mode
        - h5file, the File instance returned when filePath is opened
        """

        super(DBDoc, self).__init__()

        # The opening mode
        self.mode = mode

        # The full path of the database
        self.filepath = filepath

        # The filename of the database
        self.filename = os.path.basename(filepath)

        # The tables.File instance
        self.h5file = self.openH5File()

        # Hidden groups are used as an intermediate storage
        # in cut operations
        self.hidden_group = None

        if is_tmp_dbdoc:
            self.h5file.create_group('/', '_p_query_results',
                                     'Hide the result of queries')
            self.h5file.flush()

    def openH5File(self):
        """Open the file tied to this instance."""

        h5file = None
        # If read-write access is required then test file writability
        if self.mode != 'r' and os.path.isfile(self.filepath):
            if not os.access(self.filepath, os.W_OK):
                self.mode = 'r'
                log.warning(
                    translate('DBDoc',
                              """File access in read-write mode """
                              """is denied. It will be opened in read-only """
                              """mode.""", 'A logger error message'))

        try:
            h5file = tables.open_file(self.filepath, self.mode)
        except IOError as inst:
            log.error(translate('DBDoc', "{0}",
                                'A logger error message').format(inst))
        except:
            vitables.utils.formatExceptionInfo()
            log.error(
                translate('DBDoc',
                          "Please, if you think it is a bug, report it to "
                          "developers.", 'A logger error message'))

        return h5file

    def closeH5File(self):
        """Closes a tables.File instance."""

        try:
            self.h5file.close()
        except (tables.NodeError, OSError):
            vitables.utils.formatExceptionInfo()

    def getFileFormat(self):
        """
        The format of the database file.

        This is an accessor method intended to be used by external
        classes.

        :Returns: the format of the database file
        """

        file_format = None
        if self.h5file:
            if self.h5file._isPTFile:
                file_format = 'PyTables file'
            else:
                file_format = 'Generic HDF5 file'

        return file_format

    def get_node(self, where):
        """
        The node whose path is where.

        :Parameter where: the full path of the retrieved node
        """

        try:
            node = self.h5file.get_node(where)
            return node
        except tables.exceptions.NoSuchNodeError:
            log.error(
                translate('DBDoc',
                          """Cannot open node {0} in file {1} """,
                          'Error message').format(where, self.filepath))
            vitables.utils.formatExceptionInfo()
            return None

    def list_nodes(self):
        """:Returns: the recursive list of full nodepaths for the file"""
        return [node._v_pathname for node in self.h5file.walk_nodes('/')]

    #
    # Editing databases
    #

    def copy_file(self, dst_filepath):
        """
        Copy the contents of this file to another one.

        .. Note:: Given two open files in a ``ViTables`` session, overwriting
          one of them via ``File --> Save As...`` command may work or not. It
          depends on the operating system. On Unices it works because a process
          can delete or truncate a file open by a different process (unless
          file is blocked). On Windows it seems to work fine too.

        If the overwriting is not done via ``ViTables`` but in an interactive
        session it fails, due (probably) to `HDF5` memory protection.

        :Parameter dst_filepath: the full path of the destination file
        """

        try:
            self.h5file.copy_file(dst_filepath, overwrite=True)
        except tables.exceptions.HDF5ExtError:
            log.error(
                translate('DBDoc',
                          """Unable to save the file {0} as """
                          """{1}. Beware that only closed files can be """
                          """safely overwritten via Save As...""",
                          'A logger error message').format(self.filepath,
                                                           dst_filepath))
            vitables.utils.formatExceptionInfo()

    def createHiddenGroup(self):
        """
        Create a hidden group for storing cut nodes.
        """

        group_name = '_p_' + str(uuid.uuid4())
        self.hidden_group = '/' + group_name
        self.h5file.create_group('/', group_name, 'Hide cut nodes')
        self.h5file.flush()




#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#       Copyright (C) 2005-2007 Carabos Coop. V. All rights reserved
#       Copyright (C) 2008-2019 Vicent Mas. All rights reserved
#
#       This program is free software: you can redistribute it and/or modify
#       it under the terms of the GNU General Public License as published by
#       the Free Software Foundation, either version 3 of the License, or
#       (at your option) any later version.
#
#       This program is distributed in the hope that it will be useful,
#       but WITHOUT ANY WARRANTY; without even the implied warranty of
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#       GNU General Public License for more details.
#
#       You should have received a copy of the GNU General Public License
#       along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#       Author:  Vicent Mas - vmas@vitables.org

"""
This module defines some convenience functions for dealing with filenodes.

Efficiently managing filenodes is pretty difficult as they behave as regular
text files. By design, regular files I/O is not optimized for performing well
when the files are huge. In fact, even medium size files perform poorly. The
linecache module helps a little with this problem but at a price: it holds the
whole file in memory.
"""

import tempfile
import os

from tables.nodes import filenode

import vitables.utils as vtutils


def isFilenode(leaf):
    """Find out if PyTables node is tied to a filenode."""

    is_filenode = False
    try:
        file = leaf._v_file
        nodepath = leaf._v_pathname
        if file.get_node_attr(nodepath, 'NODE_TYPE') == filenode.NodeType:
            is_filenode = True
    except AttributeError:
        pass
    finally:
        return is_filenode


def filenodeToFile(leaf):
    """Write to disk a filenode as a regular text file."""

    # Warning: this is SLOW for large files
    fd, temp_filename = tempfile.mkstemp('.txt', 'filenode_')
    os.close(fd)
    h5file = leaf._v_file
    where = leaf._v_parent._v_pathname
    name = leaf._v_name
    filenode.read_from_filenode(h5file, temp_filename, where, name, overwrite=True)
    return vtutils.forwardPath(temp_filename)


def filenodeTotalRows(leaf):
    """Traverse the whole filenode and count its number of rows."""

    # Warning: this is SLOW for large files
    counter = 0
    with filenode.open_node(leaf, 'r') as f:
        for l in f:
            counter += 1
    return counter



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#       Copyright (C) 2008-2019 Vicent Mas. All rights reserved
#
#       This program is free software: you can redistribute it and/or modify
#       it under the terms of the GNU General Public License as published by
#       the Free Software Foundation, either version 3 of the License, or
#       (at your option) any later version.
#
#       This program is distributed in the hope that it will be useful,
#       but WITHOUT ANY WARRANTY; without even the implied warranty of
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#       GNU General Public License for more details.
#
#       You should have received a copy of the GNU General Public License
#       along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#       Author:  Vicent Mas - vmas@vitables.org

"""
This module provides a widget with the full description of the plugin.

The module allows for configuring the plugin too.

The widget is displayed when the plugin is selected in the Preferences
dialog selector pane.
"""

import os.path
try:
    import configparser
except ImportError:
    import ConfigParser as configparser

from qtpy import QtGui
from qtpy import QtWidgets
from qtpy.uic import loadUiType

__docformat__ = 'restructuredtext'

# This method of the PyQt5.uic module allows for dynamically loading user
# interfaces created by QtDesigner. See the PyQt5 Reference Guide for more
# info.
Ui_DBsTreeSortPage = \
    loadUiType(os.path.join(os.path.dirname(__file__),
                            'dbs_tree_sort_page.ui'))[0]


class AboutPage(QtWidgets.QWidget, Ui_DBsTreeSortPage):
    """
    Widget for describing and customizing the Sorting of DBs Tree plugin.

    By loading UI files at runtime we can:

        - create user interfaces at runtime (without using pyuic)
        - use multiple inheritance, MyParentClass(BaseClass, FormClass)

    This widget is inserted as a page in the stacked widget of the
    Preferences dialog when the sorting of DBs Tree item is clicked in the
    selector tree.

    """

    def __init__(self, desc, parent=None):
        """Fill and customize the widget describing the DBs sorting plugin.

        :Parameters:
          -`desc` a dictionary with the plugin description
          -`parent` the stacked widget where this About page will be added
        """

        # Makes the dialog and gives it a layout
        super(AboutPage, self).__init__(parent)
        self.setupUi(self)

        # The widget with the OK and Cancel buttons
        self.preferences_dlg = parent.parent()
        self.dlg_box_buttons = self.preferences_dlg.buttonsBox

        # The plugin description
        self.version_le.setText(desc['version'])
        self.module_name_le.setText(desc['module_name'])
        self.folder_le.setText(desc['folder'])
        self.author_le.setText(desc['author'])
        self.desc_te.setText(desc['about_text'])

        # The absolute path of the INI file
        self.ini_filename = \
            os.path.join(os.path.dirname(__file__), 'sorting_algorithm.ini')

        # Read initial configuration from the INI file
        self.config = configparser.ConfigParser()
        default_sorting = 'human'
        try:
            self.config.read_file(open(self.ini_filename))
            self.initial_sorting = self.config['DBsTreeSorting']['algorithm']
        except (IOError, configparser.ParsingError):
            self.initial_sorting = default_sorting

        # Fill the combo with values and choose the current sorting algorithm
        self.algorithms_combobox.insertItems(
            0, ['alphabetical', 'human'])
        current_index = self.algorithms_combobox.findText(self.initial_sorting)
        self.algorithms_combobox.setCurrentIndex(current_index)

        # Connect signals to slots
        self.dlg_box_buttons.button(QtWidgets.QDialogButtonBox.Cancel).clicked.\
            connect(self.cancelAlgorithmChange)
        self.dlg_box_buttons.button(QtWidgets.QDialogButtonBox.Ok).clicked.\
            connect(self.saveAlgorithmChange)

    def cancelAlgorithmChange(self):
        """Restore the initial sorting algorithm in the combobox.

        If the user press the Cancel button changes made in the
        combobox are lost.
        """

        self.config['DBsTreeSorting']['algorithm'] = self.initial_sorting
        with open(self.ini_filename, 'w') as ini_file:
            self.config.write(ini_file)
        # self.preferences_dlg.reject()

    def saveAlgorithmChange(self):
        """Save the combobox current algorithm.
        """

        self.config['DBsTreeSorting']['algorithm'] = \
            self.algorithms_combobox.currentText()
        with open(self.ini_filename, 'w') as ini_file:
            self.config.write(ini_file)
        # self.parent().parent().accept()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#       Copyright (C) 2005-2007 Carabos Coop. V. All rights reserved
#       Copyright (C) 2008-2019 Vicent Mas. All rights reserved
#
#       This program is free software: you can redistribute it and/or modify
#       it under the terms of the GNU General Public License as published by
#       the Free Software Foundation, either version 3 of the License, or
#       (at your option) any later version.
#
#       This program is distributed in the hope that it will be useful,
#       but WITHOUT ANY WARRANTY; without even the implied warranty of
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#       GNU General Public License for more details.
#
#       You should have received a copy of the GNU General Public License
#       along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#       Author:  Vicent Mas - vmas@vitables.org

"""
This module manages the ``ViTables`` configuration.

The module provides methods for reading and writing settings. Whether the
settings are stored in a plain text file or in a Windows registry is
transparent for this module because it deals with settings via
`QtCore.QSettings`.

Every access to the config settings is done via a `QSettings` instance that,
in turn, will access the config file and return the read setting to the
application. Saving settings works in a similar way, the application passes
the setting to the `QSetting` instance and it (the instance) will write the
setting into the config file.

.. Note:: *About the config file location*.
  If format is NativeFormat then the default search path will be:

  - Unix

    - UserScope

      - ``$HOME/.config/MyCompany/ViTables-version.conf``
      - ``$HOME/.config/MyCompany.conf``

    - SystemScope

      - ``/etc/xdg/MyCompany/ViTables-version.conf``
      - ``/etc/xdg/MyCompany.conf``

  - MacOSX

    - UserScope

      - ``$HOME/Library/Preferences/org.vitables.ViTables-version.plist``
      - ``$HOME/Library/Preferences/org.vitables.plist``

    - SystemScope

      - ``/Library/Preferences/org.vitables.ViTables-version.plist``
      - ``/Library/Preferences/org.vitables.plist``

  - Windows

    - UserScope

      - ``HKEY_CURRENT_USER/Software/MyCompany/ViTables/version``
      - ``HKEY_CURRENT_USER/Software/MyCompany/``

    - SystemScope

      - ``HKEY_LOCAL_MACHINE/Software/MyCompany/ViTables/version``
      - ``HKEY_LOCAL_MACHINE/Software/MyCompany/``

  If format is NativeFormat and platform is Unix the path can be set via
  QSettings.setPath static method.

.. Note:: *About the config file name*.
  If format is NativeFormat:

  - under Unix, Product Name -> Product Name.conf so the product name
    ``ViTables`` will match a configuration file named ``ViTables.conf``
  - under MacOSX, Internet Domain and Product Name ->
    reversed Internet Domain.Product Name.plist so the domain
    ``vitables.org`` and the product ``ViTables`` become
    ``org.vitables.ViTables.plist``

"""

import logging
import sys
from vitables.preferences import cfgexception
import vitables.utils

from qtpy import QtCore
from qtpy import QtGui
from qtpy import QtWidgets

import vitables.vttables.datasheet as datasheet


__docformat__ = 'restructuredtext'
__version__ = '3.0.2'

translate = QtWidgets.QApplication.translate


def getVersion():
    """The application version."""
    return __version__


log = logging.getLogger(__name__)


class Config(QtCore.QSettings):
    """
    Manages the application configuration dynamically.

    This class defines accessor methods that allow the application (a
    :meth:`vitables.vtapp.VTApp` instance) to read the configuration in
    file/registry/plist. The class also provides a method to save the current
    configuration in the configuration file/registry/plist.
    """

    def __init__(self):
        """
        Setup the application configurator.

        On Windows systems settings will be stored in the registry
        under the HKCU\Software\ViTables\__version__ key
        Mac OS X saves settings in a properties list stored in a
        standard location, either on a global or user basis (see
        docstring for more information).

        In all platforms QSettings format is NativeFormat and scope
        is UserScope.
        """

        organization = QtWidgets.qApp.organizationName()
        product = QtWidgets.qApp.applicationName()
        version = QtWidgets.qApp.applicationVersion()
        if sys.platform.startswith('win'):
            # organizationName() -> product
            # applicationName() -> version
            super(Config, self).__init__(product, version)
            self.reg_path = 'HKEY_CURRENT_USER\\Software\\{0}\\{1}'.format(
                product, version)
            self.setPath(QtCore.QSettings.NativeFormat,
                         QtCore.QSettings.UserScope, self.reg_path)
        elif sys.platform.startswith('darwin'):
            # organizationName() -> product
            # applicationName() -> version
            super(Config, self).__init__(product, version)
        else:
            # organizationName() -> organization
            # applicationName() -> product-version
            arg1 = organization
            arg2 = '-'.join((product, version))
            super(Config, self).__init__(arg1, arg2)

        # System-wide settings will not be searched as a fallback
        # Setting the NativeFormat paths on MacOSX has no effect
        self.setFallbacksEnabled(False)

        # The application default style depends on the platform
        styles = QtWidgets.QStyleFactory.keys()
        self.default_style = styles[0]
        self.vtapp = vitables.utils.getVTApp()
        if not (self.vtapp is None):
            style_name = self.vtapp.gui.style().objectName()
            for item in styles:
                if item.lower() == style_name:
                    self.default_style = item
                    break

    def loggerPaper(self):
        """
        Returns the logger background color.
        """

        key = 'Logger/Paper'
        default_value = QtGui.QColor("#ffffff")
        setting_value = self.value(key)
        if isinstance(setting_value, QtGui.QColor):
            return setting_value
        else:
            return default_value

    def loggerText(self):
        """
        Returns the logger text color.
        """

        key = 'Logger/Text'
        default_value = QtGui.QColor("#000000")
        setting_value = self.value(key)
        if isinstance(setting_value, QtGui.QColor):
            return setting_value
        else:
            return default_value

    def loggerFont(self):
        """
        Returns the logger font.
        """

        key = 'Logger/Font'
        default_value = QtWidgets.qApp.font()
        setting_value = self.value(key)
        if isinstance(setting_value, QtGui.QFont):
            return setting_value
        else:
            return default_value

    def workspaceBackground(self):
        """
        Returns the workspace background color.
        """

        key = 'Workspace/Background'
        default_value = QtGui.QBrush(QtGui.QColor("#ffffff"))
        setting_value = self.value(key)
        if isinstance(setting_value, QtGui.QBrush):
            return setting_value
        else:
            return default_value

    def readStyle(self):
        """Returns the current application style."""

        # The property key and its default value
        key = 'Look/currentStyle'
        default_value = self.default_style

        # Read the entry from the configuration file/registry
        setting_value = self.value(key)

        # Check the entry format and value
        styles = QtWidgets.QStyleFactory.keys()
        if not isinstance(setting_value, str):
            return default_value
        elif setting_value not in styles:
            return default_value
        else:
            return setting_value

    def windowPosition(self):
        """
        Returns the main window geometry settings.

        Basically the main window geometry is made of the x and y coordinates
        of the top left corner, width and height. A QByteArray with all this
        information can be created via QMainWindow.saveGeometry() and stored
        in a setting with QSetting.setValue()
        """

        key = 'Geometry/Position'
        default_value = None
        setting_value = self.value(key)
        if isinstance(setting_value, QtCore.QByteArray):
            return setting_value
        else:
            return default_value

    def windowLayout(self):
        """
        Returns the main window layout setting.

        This setting stores the position and size of toolbars and
        dockwidgets.
        """

        key = 'Geometry/Layout'
        default_value = None
        setting_value = self.value(key)
        if isinstance(setting_value, QtCore.QByteArray):
            return setting_value
        else:
            return default_value

    def hsplitterPosition(self):
        """
        Returns the horizontal splitter geometry setting.
        """

        key = 'Geometry/HSplitter'
        default_value = None
        setting_value = self.value(key)
        if isinstance(setting_value, QtCore.QByteArray):
            return setting_value
        else:
            return default_value

    def restoreLastSession(self):
        """
        Returns the `Restore last session` setting.

        This is a user preference that can be setup in the Preferences dialog,
        with the 'Restore last session' checkbox.
        """

        key = 'Session/restoreLastSession'
        default_value = False
        # Warning!
        # If the application settings have not yet been saved
        # in the registry then self.value(key) returns a Null
        # QVariant (its type is None) and self.value(key, type=bool)
        # raises an exception because None cannot be converted
        # to a boolean value
        try:
            setting_value = self.value(key, type=bool)
        except TypeError:
            setting_value = default_value
        if setting_value in (False, True):
            return setting_value
        else:
            return default_value

    def startupWorkingDir(self):
        """
        Returns the `Startup working directory` setting.

        The startup working directory is the directory to be accessed the very
        first time a file opening happens in a working session. The setting can
        have two values:
        - last -> go to the last directory accessed in the previous ViTables
            session
        - current -> go to the current working directory if ViTables started
            from a terminal and go to the home directory otherwise

        This is a user preference that can be setup in the Preferences dialog,
        with the 'Start in last working directory' checkbox.
        """

        key = 'Session/startupWorkingDir'
        default_value = 'home'
        setting_value = self.value(key)
        if setting_value in ['last', 'current']:
            return setting_value
        else:
            return default_value

    def lastWorkingDir(self):
        """
        Returns the `Last working directory` setting.
        """

        key = 'Session/lastWorkingDir'
        default_value = vitables.utils.getHomeDir()
        setting_value = self.value(key)
        if isinstance(setting_value, str):
            # TODO: check if the value is an existing file
            return setting_value
        else:
            return default_value

    def recentFiles(self):
        """
        Returns the list of most recently opened files setting.
        """

        key = 'Recent/Files'
        default_value = []
        setting_value = self.value(key)
        if isinstance(setting_value, list):
            return setting_value
        else:
            return default_value

    def sessionFiles(self):
        """
        Returns the list of files and nodes opened when the last session quit.
        """

        key = 'Session/Files'
        default_value = []
        setting_value = self.value(key)
        if isinstance(setting_value, list):
            return setting_value
        else:
            return default_value

    def helpHistory(self):
        """
        Returns the navigation history of the docs browser.
        """

        key = 'HelpBrowser/History'
        default_value = []
        setting_value = self.value(key)
        if isinstance(setting_value, list):
            return setting_value
        else:
            return default_value

    def helpBookmarks(self):
        """
        Returns the bookmarks of the docs browser.
        """

        key = 'HelpBrowser/Bookmarks'
        default_value = []
        setting_value = self.value(key)
        if isinstance(setting_value, list):
            return setting_value
        else:
            return default_value

    def enabledPlugins(self):
        """Returns the list of enabled plugins.
        """

        key = 'Plugins/Enabled'
        default_value = []
        setting_value = self.value(key)
        if isinstance(setting_value, list):
            return setting_value
        else:
            return default_value

    def writeValue(self, key, value):
        """
        Write an entry to the configuration file.

        :Parameters:

        - `key`: the name of the property we want to set.
        - `value`: the value we want to assign to the property
        """

        try:
            self.setValue(key, value)
            if self.status():
                raise cfgexception.ConfigFileIOException(
                    '{0}={1}'.format(key, value))
        except cfgexception.ConfigFileIOException as inst:
            log.error(inst.error_message)

    def readConfiguration(self):
        """
        Get the application configuration currently stored on disk.

        Read the configuration from the stored settings. If a setting
        cannot be read (as it happens when the package is just
        installed) then its default value is returned.
        Geometry and Recent settings are returned as lists, color
        settings as QColor instances. The rest of settings are returned
        as strings or integers.

        :Returns: a dictionary with the configuration stored on disk
        """

        config = {}
        config['Logger/Paper'] = self.loggerPaper()
        config['Logger/Text'] = self.loggerText()
        config['Logger/Font'] = self.loggerFont()
        config['Workspace/Background'] = self.workspaceBackground()
        config['Session/restoreLastSession'] = self.restoreLastSession()
        config['Session/startupWorkingDir'] = self.startupWorkingDir()
        config['Session/lastWorkingDir'] = self.lastWorkingDir()
        config['Geometry/Position'] = self.windowPosition()
        config['Geometry/Layout'] = self.windowLayout()
        config['Geometry/HSplitter'] = self.hsplitterPosition()
        config['Recent/Files'] = self.recentFiles()
        config['Session/Files'] = self.sessionFiles()
        config['HelpBrowser/History'] = self.helpHistory()
        config['HelpBrowser/Bookmarks'] = self.helpBookmarks()
        config['Look/currentStyle'] = self.readStyle()
        config['Plugins/Enabled'] = self.enabledPlugins()
        return config

    def applyConfiguration(self, config):
        """
        Configure the application with the given settings.

        We call `user settings` to those settings that can be setup via
        Settings dialog and `internal settings` to the rest of settings.

        At startup all settings will be loaded. At any time later the
        `users settings` can be explicitly changed via Settings dialog.

        :Parameter config: a dictionary with the settings to be (re)loaded
        """

        # Load the user settings
        self.applyUserPreferences(config)

        # Load the internal settings (if any)
        gui = self.vtapp.gui
        try:
            key = 'Geometry/Position'
            value = config[key]
            if isinstance(value, QtCore.QByteArray):
                # Default position is provided by the underlying window manager
                gui.restoreGeometry(value)

            key = 'Geometry/Layout'
            value = config[key]
            if isinstance(value, QtCore.QByteArray):
                # Default layout is provided by the underlying Qt installation
                gui.restoreState(value)

            key = 'Geometry/HSplitter'
            value = config[key]
            if isinstance(value, QtCore.QByteArray):
                # Default geometry provided by the underlying Qt installation
                gui.hsplitter.restoreState(value)

            key = 'Session/lastWorkingDir'
            self.last_working_directory = config[key]

            key = 'Recent/Files'
            self.recent_files = config[key]

            key = 'Session/Files'
            self.session_files_nodes = config[key]

            key = 'HelpBrowser/History'
            self.hb_history = config[key]

            key = 'HelpBrowser/Bookmarks'
            self.hb_bookmarks = config[key]
        except KeyError:
            pass

    def applyUserPreferences(self, config):
        """Apply settings that can be setup via Settings dialog.

        :Parameter config: a dictionary with the settings to be (re)loaded
        """

        # Usually after calling the Settings dialog only some user
        # settings will need to be reloaded. So for every user setting
        # we have to check if it needs to be reloaded or not
        key = 'Session/restoreLastSession'
        if key in config:
            self.restore_last_session = config[key]

        key = 'Session/startupWorkingDir'
        if key in config:
            self.initial_working_directory = config[key]

        key = 'Logger/Paper'
        logger = self.vtapp.gui.logger
        if key in config:
            value = config[key]
            paper = value.name()
            stylesheet = logger.styleSheet()
            old_paper = stylesheet[-7:]
            new_stylesheet = stylesheet.replace(old_paper, paper)
            logger.setStyleSheet(new_stylesheet)

        key = 'Logger/Text'
        if key in config:
            logger.moveCursor(QtGui.QTextCursor.End)
            logger.setTextColor(config[key])

        key = 'Logger/Font'
        if key in config:
            logger.setFont(config[key])

        key = 'Workspace/Background'
        workspace = self.vtapp.gui.workspace
        if key in config:
            workspace.setBackground(config[key])
            workspace.viewport().update()

        key = 'Look/currentStyle'
        if key in config:
            self.current_style = config[key]
            # Default style is provided by the underlying window manager
            QtWidgets.qApp.setStyle(self.current_style)

        key = 'Plugins/Enabled'
        if key in config:
            self.enabled_plugins = config[key]

    def saveConfiguration(self):
        """
        Store current application settings on disk.

        Note that we are using ``QSettings`` for writing to the config file,
        so we **must** rely on its searching algorithms in order to find
        that file.
        """

        vtgui = self.vtapp.gui
        # Logger paper
        style_sheet = vtgui.logger.styleSheet()
        paper = style_sheet[-7:]
        self.writeValue('Logger/Paper', QtGui.QColor(paper))
        # Logger text color
        self.writeValue('Logger/Text', vtgui.logger.textColor())
        # Logger text font
        self.writeValue('Logger/Font', vtgui.logger.font())
        # Workspace
        self.writeValue('Workspace/Background', vtgui.workspace.background())
        # Style
        self.writeValue('Look/currentStyle', self.current_style)
        # Startup working directory
        self.writeValue('Session/startupWorkingDir',
                        self.initial_working_directory)
        # Startup restore last session
        self.writeValue('Session/restoreLastSession',
                        self.restore_last_session)
        # Startup last working directory
        self.writeValue('Session/lastWorkingDir', self.last_working_directory)
        # Window geometry
        self.writeValue('Geometry/Position', vtgui.saveGeometry())
        # Window layout
        self.writeValue('Geometry/Layout', vtgui.saveState())
        # Horizontal splitter geometry
        self.writeValue('Geometry/HSplitter', vtgui.hsplitter.saveState())
        # The list of recent files
        self.writeValue('Recent/Files', self.recent_files)
        # The list of session files and nodes
        self.session_files_nodes = self.getSessionFilesNodes()
        self.writeValue('Session/Files', self.session_files_nodes)
        # The Help Browser history
        self.writeValue('HelpBrowser/History', self.hb_history)
        # The Help Browser bookmarks
        self.writeValue('HelpBrowser/Bookmarks', self.hb_bookmarks)
        # The list of enabled plugins
        self.writeValue('Plugins/Enabled',
                        self.vtapp.plugins_mgr.enabled_plugins)
        self.sync()

    def getSessionFilesNodes(self):
        """
        The list of files and nodes currently open.

        The list looks like::

            ['mode#@#filepath1#@#nodepath1#@#nodepath2, ...',
            'mode#@#filepath2#@#nodepath1#@#nodepath2, ...', ...]
        """

        # Get the list of views
        workspace = self.vtapp.gui.workspace
        node_views = [window for window in workspace.subWindowList()
                      if isinstance(window, datasheet.DataSheet)]

        # Get the list of open files (temporary database is not included)
        dbt_model = self.vtapp.gui.dbs_tree_model
        session_files_nodes = []
        filepaths = dbt_model.getDBList()

        for path in filepaths:
            mode = dbt_model.getDBDoc(path).mode
            # If a new file has been created during the current session
            # then write mode must be replaced by append mode or the file
            # will be created from scratch in the next ViTables session
            if mode == 'w':
                mode = 'a'
            item_path = mode + '#@#' + path
            for view in node_views:
                if view.dbt_leaf.filepath == path:
                    item_path = item_path + '#@#' + view.dbt_leaf.nodepath
            session_files_nodes.append(item_path)

        # Format the list in a handy way to store it on disk
        return session_files_nodes



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#       Copyright (C) 2005-2007 Carabos Coop. V. All rights reserved
#       Copyright (C) 2008-2019 Vicent Mas. All rights reserved
#
#       This program is free software: you can redistribute it and/or modify
#       it under the terms of the GNU General Public License as published by
#       the Free Software Foundation, either version 3 of the License, or
#       (at your option) any later version.
#
#       This program is distributed in the hope that it will be useful,
#       but WITHOUT ANY WARRANTY; without even the implied warranty of
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#       GNU General Public License for more details.
#
#       You should have received a copy of the GNU General Public License
#       along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#       Author:  Vicent Mas - vmas@vitables.org

"""
This module defines a data structure to be used for the model of the databases
tree. The data structure is equivalent to a (non root) group node in a
`PyTables` file.
"""

__docformat__ = 'restructuredtext'

import vitables.utils
from vitables.h5db import tnode_editor
from vitables.nodeprops import nodeinfo
from vitables.nodeprops import grouppropdlg

class GroupNode(object):
    """
    A group node in the tree of databases model.

    :Parameters:

    - `parent`: the parent of the node.
    - `name`: the name of the node
    """

    def __init__(self, model, parent, name):
        """Create a group node for the tree of databases model.

        A GroupNode represents a (non root) group of a `HDF5` file and has
        a parent (another group node of the tree of databases model) and
        a name.
        """

        self.dbt_model = model
        self.updated = False
        self.children = []
        self.parent = parent
        self.node = parent.node._f_get_child(name)
        self.node_kind = 'group'

        self.has_view = False

        # Attributes that the tree of databases view will use
        # name --> DisplayRole
        # nodepath --> ToolTipRole
        # as_record --> StatusTipRole
        # icon --> DecorationRole
        self.name = name
        parentpath = parent.nodepath
        if parentpath.endswith('/'):
            parentpath = parentpath[:-1]
        self.nodepath = '{0}/{1}'.format(parentpath, name)
        self.filepath = parent.filepath
        self.as_record = '{0}->{1}'.format(self.filepath, self.nodepath)
        icons = vitables.utils.getIcons()
        self.closed_folder = icons['folder']
        self.open_folder = icons['document-open-folder']
        self.icon = icons['folder']


    def __len__(self):
        """The number of children of this grup."""
        return len(self.children)


    def insertChild(self, child, position=0):
        """Insert a child in a group node.

        :Parameters:

            - `child`: the node being inserted
            - `position`: the insertion position
        """
        self.children.insert(position, child)


    def childAtRow(self, row):
        """The row-th child of this node.

        :Parameter row: the position of the retrieved child
        """

        assert 0 <= row <= len(self.children)
        return self.children[row]


    def rowOfChild(self, child):
        """The row index of a given child.

        :Parameter child: the child node whose position is being retrieved.
        """

        for pos, node in enumerate(self.children):
            if node == child:
                return pos
        return -1


    def row(self):
        """The position of this node in the parent's list of children.
        """

        if self.parent:
            return self.parent.children.index(self)

        return 0


    def findChild(self, childname):
        """The child node with a given name.

        :Parameter childname: the name of the wanted child node.
        """

        for node in self.children:
            if node.name == childname:
                return node
        return None


    def editor(self):
        """Return an instance of `TNodeEditor`.
        """
        return tnode_editor.TNodeEditor(self.dbt_model.getDBDoc(self.filepath))


    def properties(self):
        """The Properties dialog for this node.
        """

        info = nodeinfo.NodeInfo(self)
        grouppropdlg.GroupPropDlg(info)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#       Copyright (C) 2005-2007 Carabos Coop. V. All rights reserved
#       Copyright (C) 2008-2019 Vicent Mas. All rights reserved
#
#       This program is free software: you can redistribute it and/or modify
#       it under the terms of the GNU General Public License as published by
#       the Free Software Foundation, either version 3 of the License, or
#       (at your option) any later version.
#
#       This program is distributed in the hope that it will be useful,
#       but WITHOUT ANY WARRANTY; without even the implied warranty of
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#       GNU General Public License for more details.
#
#       You should have received a copy of the GNU General Public License
#       along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#       Author:  Vicent Mas - vmas@vitables.org

"""Create a PyTables file for testing purposes."""

from tables.nodes import filenode
import tables


class Particle(tables.IsDescription):
    """Description of a table record.
    """
    name        = tables.StringCol(16, pos=1)   # 16-character String
    lati        = tables.Int32Col(pos=2)        # integer
    longi       = tables.Int32Col(pos=3)        # integer
    pressure    = tables.Float32Col(pos=4)      # float  (single-precision)
    temperature = tables.Float64Col(pos=5)      # double (double-precision)

# Create a file in write mode
h5file = tables.open_file('testfile.h5', 'w')

# Create a table
root = h5file.root
group = h5file.create_group(root, "tables")
table = h5file.create_table(group, 'Particles', Particle, "A table", tables.Filters(1))

# Fill the table with 10 particles
particle = table.row
for i in range(10):
    # First, assign the values to the Particle record
    particle['name']  = 'Particle: {:6d}'.format(i)
    particle['lati'] = i
    particle['longi'] = 10 - i
    particle['pressure'] = float(i*i)
    particle['temperature'] = float(i**2)
    # This injects the row values.
    particle.append()

# We need to flush the buffers in table in order to get an
# accurate number of records on it.
table.flush()

# Create a filenode
fnode = filenode.new_node(h5file, where='/', name='filenode')

# Fill the filenode
counter = 0
while counter < 10:
    l = "This is a line inserted programmatically at position {}\n".format(counter)
    fnode.write(l.encode("utf-8"))
    counter += 1
fnode.write(bytearray("This is the last line.\n", "utf-8"))
fnode.attrs.author = "Vicent Mas"
fnode.close()

h5file.close()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#       Copyright (C) 2005-2007 Carabos Coop. V. All rights reserved
#       Copyright (C) 2008-2019 Vicent Mas. All rights reserved
#
#       This program is free software: you can redistribute it and/or modify
#       it under the terms of the GNU General Public License as published by
#       the Free Software Foundation, either version 3 of the License, or
#       (at your option) any later version.
#
#       This program is distributed in the hope that it will be useful,
#       but WITHOUT ANY WARRANTY; without even the implied warranty of
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#       GNU General Public License for more details.
#
#       You should have received a copy of the GNU General Public License
#       along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#       Author:  Vicent Mas - vmas@vitables.org

"""
The controller of the Users' Guide browser is implemented in this module.

The documentation browser is used for browsing the HTML version of the Guide
which is distributed along the source code in every supported platform.
"""

__docformat__ = 'restructuredtext'

from qtpy import QtCore
from qtpy import QtGui
from qtpy import QtWidgets

import vitables.utils
from vitables.docbrowser import bookmarksdlg
from vitables.docbrowser import browsergui

translate = QtWidgets.QApplication.translate


class HelpBrowser(QtCore.QObject) :
    """
    This is the controller of the ``ViTables`` documentation browser.
    Features:

    * reasonably good understanding of ``HTML`` format
    * navigation toolbar
    * session capable

      - filenames of files opened during the session are available at
          any moment
      - last session can be restored next time we open the `HelpBrowser`

    * basic bookmarks management
    """

    def __init__(self) :
        """
        Initializes the browser.

        Restore history and bookmarks from disk and makes the browser window.
        """

        super(HelpBrowser, self).__init__()

        self.vtapp = vitables.utils.getVTApp()

        # Get the bookmarks and the navigation history
        self.bookmarks = self.vtapp.config.hb_bookmarks
        self.history = self.vtapp.config.hb_history

        # create the GUI
        self.gui = browsergui.HelpBrowserGUI(self)

        # The working directory used in QFileDialog calls
        self.working_dir = vitables.utils.getHomeDir()

        # The GUI setup is slow so it is not shown until the setup is
        # done (it avoids displaying an ugly empty widget)
        self.displaySrc('index.html')
        self.gui.show()


    def displaySrc(self, src=False):
        """
        Displays a document in the `HelpBrowser` window.

        This slot is called when:

            - :meth:
                `vitables.docbrowser.bookmarksdlg.BookmarksDlg.displayBookmark`
                is launched
            - a new item is activated in the `History` combo
            - an entry is selected in the `Bookmarks` menu

        :Parameter src: the path of the file being displayed
        """

        if src is False: # entry selected in the Bookmarks menu
            action = self.gui.sender()
            src = action.data()

        src = QtCore.QDir(src).dirName()
        src = QtCore.QDir().fromNativeSeparators(src)
        url = QtCore.QUrl(src)
        self.gui.text_browser.setSource(url)

    #########################################################
    #
    # 					File menu related slots
    #
    #########################################################

    def exitBrowser(self) :
        """
        Quit the `HelpBrowser`.

        Before quitting this slot saves session and bookmarks. Then all browser
        windows are closed. The saved state corresponds to the window on which
        Exit is invoqued.
        """

        # Close all browsers
        self.vtapp.config.hb_history = self.history
        self.vtapp.config.hb_bookmarks = self.bookmarks
        self.vtapp.doc_browser = None
        self.gui.close()

    #########################################################
    #
    # 					View menu related slots
    #
    #########################################################

    def zoomIn(self) :
        """Increases the font size of the displayed source."""
        self.gui.text_browser.zoomIn(2)


    def zoomOut(self) :
        """Increases the font size of the displayed source."""
        self.gui.text_browser.zoomOut(2)

    #########################################################
    #
    # 					Go menu related slots
    #
    #########################################################

    def updateHome(self) :
        """
        Enables/disables the ``Go --> Home`` menu item.

        If the history list is empty the menu item is disabled. Otherwise
        it is enabled.
        """

        if self.gui.text_browser.source():
            self.gui.actions['goHome'].setEnabled(1)
        else:
            self.gui.actions['goHome'].setEnabled(0)


    def updateForward(self, available) :
        """Enables/disables the ``Go --> Forward`` menu item.

        :Parameter available: a boolean. Indicates if a Next page is available.
        """
        self.gui.actions['goForward'].setEnabled(available)


    def updateBackward(self, available) :
        """Enables/disables the ``Go --> Backward`` menu item.

        :Parameter available: a boolean. Indicates if a Previous page is
            available
        """
        self.gui.actions['goBackward'].setEnabled(available)

    #########################################################
    #
    # 					Bookmarks menu related slots
    #
    #########################################################

    def addBookmark(self) :
        """
        Add the current page to the bookmarks menu.
        """

        src = self.gui.text_browser.source().toString()
        src = QtCore.QDir.fromNativeSeparators(src)
        src = src.replace('///', '/')
        if self.bookmarks.count(src) :
            # if the page is already bookmarked we do nothing
            pass
        else :
            # The page is not bookmarked so we append it to the bookmarks list.
            self.bookmarks.append(src)


    def editBookmarks(self) :
        """
        Raise the dialog for editing bookmarks.
        """

        # update bookmarks list
        edit_dlg = bookmarksdlg.BookmarksDlg(self.bookmarks, self.gui)
        edit_dlg.exec_()


    def clearBookmarks(self) :
        """Clear all bookmarks."""
        self.bookmarks = []

    #########################################################
    #
    # 					History related slots
    #
    #########################################################

    def clearHistory(self) :
        """
        Clear the history of visited pages.
        """

        self.history = []
        self.gui.combo_history.clear()
        self.updateHome()


    def updateHistory(self, src) :
        """
        Updates the contents of the combo selector.

        Combo selector has to be updated when:

        - a new file is opened
        - a link is clicked in the current document
        - a bookmark is opened

        The history combobox is a visual representation of the `history list`,
        so entries will occupy the same position in both, the list and the
        combo.

        :Parameter src: the path being added to the combo
        """

        url = QtCore.QDir.fromNativeSeparators(src.toString())
        url = url.replace('///', '/')
        if url not in self.history:
            self.history.append(url)
            self.gui.combo_history.addItem(url)

        self.gui.combo_history.setCurrentIndex(self.history.index(url))
        self.updateHome()

    #########################################################
    #
    # 					Help menu related slots
    #
    #########################################################

    def aboutBrowser(self) :
        """
        Shows a message box with the application `About` info.
        """

        QtWidgets.QMessageBox.information(self.gui,
            translate('HelpBrowser', 'About HelpBrowser', 'A dialog caption'),
            translate('HelpBrowser', """<html><h3>HelpBrowser</h3>
                HelpBrowser is a very simple tool for displaying the HTML
                version of the ViTables' Guide without using external programs.
                <p>Best of all... it is written using PyQt, the Python
                bindings for the Qt GUI toolkit.</html>""",
                'About Help browser text')
            )


    def aboutQt(self) :
        """
        Shows a message box with the `Qt About` info.
        """

        caption = translate('HelpBrowser', 'About Qt', 'A dialog caption')
        QtWidgets.QMessageBox.aboutQt(self.gui, caption)



"""Default plugin about page."""

import os

from qtpy import QtGui
from qtpy import QtWidgets
from qtpy import uic

Ui_AboutPage = uic.loadUiType(os.path.join(os.path.dirname(
    os.path.abspath(__file__)), 'about_page.ui'))[0]


class AboutPage(QtWidgets.QWidget, Ui_AboutPage):
    def __init__(self, desc, parent=None):
        super(AboutPage, self).__init__(parent)
        self.setupUi(self)
        # fill gui elements
        self.version_text.setText(desc.get('version', ''))
        self.module_name_text.setText(desc.get('module_name', ''))
        self.folder_text.setText(desc.get('folder', ''))
        self.author_text.setText(desc.get('author', ''))
        self.desc_text.setText(desc.get('comment', ''))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#       Copyright (C) 2005-2007 Carabos Coop. V. All rights reserved
#       Copyright (C) 2008-2019 Vicent Mas. All rights reserved
#
#       This program is free software: you can redistribute it and/or modify
#       it under the terms of the GNU General Public License as published by
#       the Free Software Foundation, either version 3 of the License, or
#       (at your option) any later version.
#
#       This program is distributed in the hope that it will be useful,
#       but WITHOUT ANY WARRANTY; without even the implied warranty of
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#       GNU General Public License for more details.
#
#       You should have received a copy of the GNU General Public License
#       along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#       Author:  Vicent Mas - vmas@vitables.org

"""
This module displays in a dialog the link information collected by the
:mod:`vitables.nodeprops.nodeinfo` module.
"""

import os.path

from qtpy import QtGui
from qtpy import QtWidgets

from qtpy.uic import loadUiType

import vitables.utils

__docformat__ = 'restructuredtext'

translate = QtWidgets.QApplication.translate
# This method of the PyQt5.uic module allows for dinamically loading user
# interfaces created by QtDesigner. See the PyQt5 Reference Guide for more
# info.
Ui_LinkPropDialog = \
    loadUiType(os.path.join(os.path.dirname(__file__), 'link_prop_dlg.ui'))[0]


class LinkPropDlg(QtWidgets.QDialog, Ui_LinkPropDialog):
    """
    Link properties dialog.

    By loading UI files at runtime we can:

        - create user interfaces at runtime (without using pyuic)
        - use multiple inheritance, MyParentClass(BaseClass, FormClass)

    This class displays a simple dialog that shows some properties of
    the selected link: name, path, type and target.

    :Parameter info: a :meth:`vitables.nodeprops.nodeinfo.NodeInfo` instance
      describing a given node
    """

    def __init__(self, info):
        """Setup the Properties dialog."""

        vtapp = vitables.utils.getVTApp()
        super(LinkPropDlg, self).__init__(vtapp.gui)
        self.setupUi(self)

        self.setWindowTitle(translate('LinkPropDlg', 'Link properties',
                                      'Dlg caption'))

        # Customise the dialog's pages
        self.fillPage(info)
        self.resize(self.size().width(), self.minimumHeight())

        # Show the dialog
        self.show()

    def fillPage(self, info):
        """Fill the dialog with info regarding the given link.

        :Parameter info: a :meth:`vitables.nodeprops.nodeinfo.NodeInfo` instance
          describing a given node
        """

        self.nameLE.setText(info.nodename)
        self.pathLE.setText(info.nodepath)
        self.typeLE.setText(info.link_type)
        self.targetLE.setText(info.target)
